import pytest

from SAF.features.coordination.helper import (
    _compute_global_avg_for_min_max_avg_metrics,
    _compute_min_max_avg_per_atomic_label,
    _compute_number_of_atoms_in_binary_CN,
    get_CN_metrics_per_method,
)


@pytest.fixture(scope="function")
def RhSb2_CN_metrics_per_method():
    return {
        "SbI": {
            "dist_by_shortest_dist": {
                "volume_of_polyhedron": 7.841,
                "distance_from_avg_point_to_center": 0.329,
                "number_of_vertices": 4,
                "number_of_edges": 6,
                "number_of_faces": 4,
                "shortest_distance_to_face": 0.69,
                "shortest_distance_to_edge": 1.078,
                "volume_of_inscribed_sphere": 1.374,
                "packing_efficiency": 0.175,
            },
            "dist_by_CIF_radius_sum": {
                "volume_of_polyhedron": 7.841,
                "distance_from_avg_point_to_center": 0.329,
                "number_of_vertices": 4,
                "number_of_edges": 6,
                "number_of_faces": 4,
                "shortest_distance_to_face": 0.69,
                "shortest_distance_to_edge": 1.078,
                "volume_of_inscribed_sphere": 1.374,
                "packing_efficiency": 0.175,
            },
            "dist_by_CIF_radius_refined_sum": {
                "volume_of_polyhedron": 7.841,
                "distance_from_avg_point_to_center": 0.329,
                "number_of_vertices": 4,
                "number_of_edges": 6,
                "number_of_faces": 4,
                "shortest_distance_to_face": 0.69,
                "shortest_distance_to_edge": 1.078,
                "volume_of_inscribed_sphere": 1.374,
                "packing_efficiency": 0.175,
            },
            "dist_by_Pauling_radius_sum": {
                "volume_of_polyhedron": 93.866,
                "distance_from_avg_point_to_center": 0.48,
                "number_of_vertices": 14,
                "number_of_edges": 33,
                "number_of_faces": 22,
                "shortest_distance_to_face": 2.127,
                "shortest_distance_to_edge": 2.146,
                "volume_of_inscribed_sphere": 40.295,
                "packing_efficiency": 0.429,
            },
        },
        "Rh": {
            "dist_by_shortest_dist": {
                "volume_of_polyhedron": 28.667,
                "distance_from_avg_point_to_center": 0.268,
                "number_of_vertices": 7,
                "number_of_edges": 15,
                "number_of_faces": 10,
                "shortest_distance_to_face": 1.509,
                "shortest_distance_to_edge": 1.749,
                "volume_of_inscribed_sphere": 14.39,
                "packing_efficiency": 0.502,
            },
            "dist_by_CIF_radius_sum": {
                "volume_of_polyhedron": 28.667,
                "distance_from_avg_point_to_center": 0.268,
                "number_of_vertices": 7,
                "number_of_edges": 15,
                "number_of_faces": 10,
                "shortest_distance_to_face": 1.509,
                "shortest_distance_to_edge": 1.749,
                "volume_of_inscribed_sphere": 14.39,
                "packing_efficiency": 0.502,
            },
            "dist_by_CIF_radius_refined_sum": {
                "volume_of_polyhedron": 28.667,
                "distance_from_avg_point_to_center": 0.268,
                "number_of_vertices": 7,
                "number_of_edges": 15,
                "number_of_faces": 10,
                "shortest_distance_to_face": 1.509,
                "shortest_distance_to_edge": 1.749,
                "volume_of_inscribed_sphere": 14.39,
                "packing_efficiency": 0.502,
            },
            "dist_by_Pauling_radius_sum": {
                "volume_of_polyhedron": 28.667,
                "distance_from_avg_point_to_center": 0.268,
                "number_of_vertices": 7,
                "number_of_edges": 15,
                "number_of_faces": 10,
                "shortest_distance_to_face": 1.509,
                "shortest_distance_to_edge": 1.749,
                "volume_of_inscribed_sphere": 14.39,
                "packing_efficiency": 0.502,
            },
        },
        "SbII": {
            "dist_by_shortest_dist": {
                "volume_of_polyhedron": 9.762,
                "distance_from_avg_point_to_center": 0.284,
                "number_of_vertices": 4,
                "number_of_edges": 6,
                "number_of_faces": 4,
                "shortest_distance_to_face": 0.733,
                "shortest_distance_to_edge": 1.257,
                "volume_of_inscribed_sphere": 1.65,
                "packing_efficiency": 0.169,
            },
            "dist_by_CIF_radius_sum": {
                "volume_of_polyhedron": 9.762,
                "distance_from_avg_point_to_center": 0.284,
                "number_of_vertices": 4,
                "number_of_edges": 6,
                "number_of_faces": 4,
                "shortest_distance_to_face": 0.733,
                "shortest_distance_to_edge": 1.257,
                "volume_of_inscribed_sphere": 1.65,
                "packing_efficiency": 0.169,
            },
            "dist_by_CIF_radius_refined_sum": {
                "volume_of_polyhedron": 9.762,
                "distance_from_avg_point_to_center": 0.284,
                "number_of_vertices": 4,
                "number_of_edges": 6,
                "number_of_faces": 4,
                "shortest_distance_to_face": 0.733,
                "shortest_distance_to_edge": 1.257,
                "volume_of_inscribed_sphere": 1.65,
                "packing_efficiency": 0.169,
            },
            "dist_by_Pauling_radius_sum": {
                "volume_of_polyhedron": 107.745,
                "distance_from_avg_point_to_center": 0.419,
                "number_of_vertices": 15,
                "number_of_edges": 36,
                "number_of_faces": 24,
                "shortest_distance_to_face": 2.391,
                "shortest_distance_to_edge": 2.179,
                "volume_of_inscribed_sphere": 57.222,
                "packing_efficiency": 0.531,
            },
        },
    }


@pytest.fixture(scope="function")
def RhSb2_min_max_avg_CN_metrics_from_sites():
    return {
        "SbI": {
            "volume_of_polyhedron": {
                "min": 7.841,
                "max": 93.866,
                "avg": 29.34725,
            },
            "distance_from_avg_point_to_center": {
                "min": 0.329,
                "max": 0.48,
                "avg": 0.36675,
            },
            "number_of_vertices": {"min": 4, "max": 14, "avg": 6.5},
            "number_of_edges": {"min": 6, "max": 33, "avg": 12.75},
            "number_of_faces": {"min": 4, "max": 22, "avg": 8.5},
            "shortest_distance_to_face": {
                "min": 0.69,
                "max": 2.127,
                "avg": 1.0492499999999998,
            },
            "shortest_distance_to_edge": {
                "min": 1.078,
                "max": 2.146,
                "avg": 1.345,
            },
            "volume_of_inscribed_sphere": {
                "min": 1.374,
                "max": 40.295,
                "avg": 11.10425,
            },
            "packing_efficiency": {"min": 0.175, "max": 0.429, "avg": 0.2385},
        },
        "Rh": {
            "volume_of_polyhedron": {
                "min": 28.667,
                "max": 28.667,
                "avg": 28.667,
            },
            "distance_from_avg_point_to_center": {
                "min": 0.268,
                "max": 0.268,
                "avg": 0.268,
            },
            "number_of_vertices": {"min": 7, "max": 7, "avg": 7.0},
            "number_of_edges": {"min": 15, "max": 15, "avg": 15.0},
            "number_of_faces": {"min": 10, "max": 10, "avg": 10.0},
            "shortest_distance_to_face": {
                "min": 1.509,
                "max": 1.509,
                "avg": 1.509,
            },
            "shortest_distance_to_edge": {
                "min": 1.749,
                "max": 1.749,
                "avg": 1.749,
            },
            "volume_of_inscribed_sphere": {
                "min": 14.39,
                "max": 14.39,
                "avg": 14.39,
            },
            "packing_efficiency": {"min": 0.502, "max": 0.502, "avg": 0.502},
        },
        "SbII": {
            "volume_of_polyhedron": {
                "min": 9.762,
                "max": 107.745,
                "avg": 34.25775,
            },
            "distance_from_avg_point_to_center": {
                "min": 0.284,
                "max": 0.419,
                "avg": 0.31775,
            },
            "number_of_vertices": {"min": 4, "max": 15, "avg": 6.75},
            "number_of_edges": {"min": 6, "max": 36, "avg": 13.5},
            "number_of_faces": {"min": 4, "max": 24, "avg": 9.0},
            "shortest_distance_to_face": {
                "min": 0.733,
                "max": 2.391,
                "avg": 1.1475,
            },
            "shortest_distance_to_edge": {
                "min": 1.257,
                "max": 2.179,
                "avg": 1.4874999999999998,
            },
            "volume_of_inscribed_sphere": {
                "min": 1.65,
                "max": 57.222,
                "avg": 15.543000000000001,
            },
            "packing_efficiency": {"min": 0.169, "max": 0.531, "avg": 0.2595},
        },
    }


def test_find_min_max_avg_CN_metrics(RhSb2_CN_metrics_per_method, RhSb2_min_max_avg_CN_metrics_from_sites):
    """Compute the min, max, and avg of the CN metrics from each site
    label across 4 methods."""
    result = _compute_min_max_avg_per_atomic_label(RhSb2_CN_metrics_per_method)
    for label in RhSb2_min_max_avg_CN_metrics_from_sites:
        for key, metric in RhSb2_min_max_avg_CN_metrics_from_sites[label].items():
            assert result[label][key]["min"] == pytest.approx(metric["min"], abs=0.005), f"{label}-{key} min is incorrect"
            assert result[label][key]["max"] == pytest.approx(metric["max"], abs=0.005), f"{label}-{key} max is incorrect"
            assert result[label][key]["avg"] == pytest.approx(metric["avg"], abs=0.005), f"{label}-{key} avg is incorrect"


def test_compute_global_averages(RhSb2_min_max_avg_CN_metrics_from_sites):
    result = _compute_global_avg_for_min_max_avg_metrics(RhSb2_min_max_avg_CN_metrics_from_sites)
    expected = {
        "min": {
            "volume_of_polyhedron": 15.423333333333334,
            "distance_from_avg_point_to_center": 0.2936666666666667,
            "number_of_vertices": 5.0,
            "number_of_edges": 9.0,
            "number_of_faces": 6.0,
            "shortest_distance_to_face": 0.9773333333333333,
            "shortest_distance_to_edge": 1.3613333333333333,
            "volume_of_inscribed_sphere": 5.804666666666667,
            "packing_efficiency": 0.28200000000000003,
        },
        "max": {
            "volume_of_polyhedron": 76.75933333333334,
            "distance_from_avg_point_to_center": 0.389,
            "number_of_vertices": 12.0,
            "number_of_edges": 28.0,
            "number_of_faces": 18.666666666666668,
            "shortest_distance_to_face": 2.009,
            "shortest_distance_to_edge": 2.0246666666666666,
            "volume_of_inscribed_sphere": 37.30233333333334,
            "packing_efficiency": 0.4873333333333334,
        },
        "avg": {
            "volume_of_polyhedron": 30.757333333333335,
            "distance_from_avg_point_to_center": 0.3175,
            "number_of_vertices": 6.75,
            "number_of_edges": 13.75,
            "number_of_faces": 9.166666666666666,
            "shortest_distance_to_face": 1.23525,
            "shortest_distance_to_edge": 1.5271666666666668,
            "volume_of_inscribed_sphere": 13.679083333333333,
            "packing_efficiency": 0.3333333333333333,
        },
    }
    # Loop through each category (min, max, avg) and each metric
    for category in ["min", "max", "avg"]:
        for metric, value in expected[category].items():
            assert result[category][metric] == pytest.approx(value, abs=0.005)


def test_compute_number_of_atoms_in_CN(RhSb2_cif, RhSb_elements_tuple):
    CN_data = get_CN_metrics_per_method(RhSb2_cif)
    A, B = RhSb_elements_tuple
    result = _compute_number_of_atoms_in_binary_CN(RhSb2_cif.connections, CN_data, A, B)
    assert result == {
        "SbI": {
            "dist_by_shortest_dist": {"A_count": 3, "B_count": 1},
            "dist_by_CIF_radius_sum": {"A_count": 3, "B_count": 1},
            "dist_by_CIF_radius_refined_sum": {"A_count": 3, "B_count": 1},
            "dist_by_Pauling_radius_sum": {"A_count": 3, "B_count": 11},
        },
        "Rh": {
            "dist_by_shortest_dist": {"A_count": 1, "B_count": 6},
            "dist_by_CIF_radius_sum": {"A_count": 1, "B_count": 6},
            "dist_by_CIF_radius_refined_sum": {"A_count": 1, "B_count": 6},
            "dist_by_Pauling_radius_sum": {"A_count": 1, "B_count": 6},
        },
        "SbII": {
            "dist_by_shortest_dist": {"A_count": 3, "B_count": 1},
            "dist_by_CIF_radius_sum": {"A_count": 3, "B_count": 1},
            "dist_by_CIF_radius_refined_sum": {"A_count": 3, "B_count": 1},
            "dist_by_Pauling_radius_sum": {"A_count": 3, "B_count": 12},
        },
    }
    min_max_avg_result = _compute_min_max_avg_per_atomic_label(result)
    assert min_max_avg_result == {
        "SbI": {
            "A_count": {"min": 3, "max": 3, "avg": 3.0},
            "B_count": {"min": 1, "max": 11, "avg": 3.5},
        },
        "Rh": {
            "A_count": {"min": 1, "max": 1, "avg": 1.0},
            "B_count": {"min": 6, "max": 6, "avg": 6.0},
        },
        "SbII": {
            "A_count": {"min": 3, "max": 3, "avg": 3.0},
            "B_count": {"min": 1, "max": 12, "avg": 3.75},
        },
    }


def test_compute_global_average_for_atom_count_in_CN():
    min_max_avg_result = {
        "SbI": {
            "A_count": {"min": 3, "max": 3, "avg": 3.0},
            "B_count": {"min": 1, "max": 11, "avg": 3.5},
        },
        "Rh": {
            "A_count": {"min": 1, "max": 1, "avg": 1.0},
            "B_count": {"min": 6, "max": 6, "avg": 6.0},
        },
        "SbII": {
            "A_count": {"min": 3, "max": 3, "avg": 3.0},
            "B_count": {"min": 1, "max": 12, "avg": 3.75},
        },
    }
    result = _compute_global_avg_for_min_max_avg_metrics(min_max_avg_result)
    expected = {
        "min": {"A_count": 2.3333333333333335, "B_count": 2.6666666666666665},
        "max": {"A_count": 2.3333333333333335, "B_count": 9.666666666666666},
        "avg": {"A_count": 2.3333333333333335, "B_count": 4.416666666666667},
    }
    for category in ["min", "max", "avg"]:
        for metric, value in expected[category].items():
            assert result[category][metric] == pytest.approx(value, abs=0.005)



import pytest

from SAF.features.interatomic.binary import compute_features


def test_binary_test(ThSb_cif):
    elements = ["Th", "Sb"]
    actual, _ = compute_features(ThSb_cif, True, elements)
    expected = {
        "Entry": "ThSb",
        "Formula": "ThSb",
        "Structure": "CsCl",
        "A": "Th",
        "B": "Sb",
        "INT_AA_dist": 3.847,
        "INT_BB_dist": 3.847,
        "INT_AB_dist": 3.332,
        "INT_Asize": 1.798,
        "INT_Bsize": 1.434,
        "INT_Asize_by_Bsize": 1.2538354253835426,
        "INT_AA_dist_by_2_by_Asize": 1.0697997775305894,
        "INT_BB_dist_by_2_by_Bsize": 1.341352859135286,
        "INT_AB_dist_by_ABsizes": 1.0309405940594059,
        "INT_Asize_ref": 1.8591213139280958,
        "INT_Bsize_ref": 1.4728786860719039,
        "INT_percent_diff_A_by_100": 0.03399405668970841,
        "INT_percent_diff_B_by_100": 0.027112054443447647,
        "INT_dist_AA_minus_ref_diff": 0.033469553455629956,
        "INT_dist_BB_minus_ref_diff": 0.2342715435030393,
        "INT_dist_AB_minus_ref_diff": 0.0,
        "INT_R_factor": 0.0018906593863675784,
        "INT_UNI_shortest_homoatomic_dist": 3.847,
        "INT_UNI_shortest_heteroatomic_dist": 3.332,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": 1.0697997775305894,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": 1.0309405940594059,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": 1.0346285557535135,
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": 1.0,
        "INT_UNI_highest_refined_percent_diff_abs": 0.03399405668970841,
        "INT_UNI_lowest_refined_percent_diff_abs": 0.027112054443447647,
        "INT_UNI_refined_packing_efficiency": 0.70785,
    }

    for key, expected_value in expected.items():
        # For float
        if isinstance(expected_value, float):
            assert actual[key] == pytest.approx(expected_value, abs=0.001)
        # For str, integer, etc.
        else:
            assert actual[key] == expected_value



from cifkit import Cif

from SAF.features.wyc.helper import _parse_site_data, get_binary_site_info, get_ternary_site_info


def test__parse_site_data():
    file_path = "tests/cif/binary/Th7Rh3.cif"
    cif = Cif(file_path)
    atomic_env = _parse_site_data(cif._loop_values)

    assert atomic_env == {
        "Rh": {
            "sites": 1,
            "multiplicity": 6,
            "lowest_wyckoff_multiplicity": 6,
            "lowest_wyckoff_element": "Rh1",
        },
        "Th": {
            "sites": 3,
            "multiplicity": 14,
            "lowest_wyckoff_multiplicity": 2,
            "lowest_wyckoff_element": "Th3",
        },
    }


def test_get_binary_site_info():
    file_path = "tests/cif/binary/Th7Rh3.cif"
    cif = Cif(file_path)
    A_info, B_info = get_binary_site_info(cif._loop_values, "Th", "Rh")

    assert A_info == {
        "sites": 3,
        "multiplicity": 14,
        "lowest_wyckoff_multiplicity": 2,
        "lowest_wyckoff_element": "Th3",
    }
    assert B_info == {
        "sites": 1,
        "multiplicity": 6,
        "lowest_wyckoff_multiplicity": 6,
        "lowest_wyckoff_element": "Rh1",
    }


def test_get_ternary_site_info():
    file_path = "tests/cif/ternary/URhIn.cif"
    cif = Cif(file_path)
    R_env, M_env, X_env = get_ternary_site_info(cif._loop_values, "U", "Rh", "In")

    assert R_env == {
        "sites": 1,
        "multiplicity": 3,
        "lowest_wyckoff_multiplicity": 3,
        "lowest_wyckoff_element": "U1",
    }
    assert M_env == {
        "sites": 2,
        "multiplicity": 3,
        "lowest_wyckoff_multiplicity": 1,
        "lowest_wyckoff_element": "Rh2",
    }

    assert X_env == {
        "sites": 1,
        "multiplicity": 3,
        "lowest_wyckoff_multiplicity": 3,
        "lowest_wyckoff_element": "In1",
    }



from cifkit import Cif

from SAF.features.wyc import helper


def compute_features(cif: Cif, elements):
    A, B = elements
    A_env, B_env = helper.get_binary_site_info(cif._loop_values, A, B)
    A_sites_total = A_env["sites"]
    B_sites_total = B_env["sites"]
    A_multiplicity_total = A_env["multiplicity"]
    B_multiplicity_total = B_env["multiplicity"]
    A_lowest_wyckoff_multiplicity = A_env["lowest_wyckoff_multiplicity"]
    B_lowest_wyckoff_multiplicity = B_env["lowest_wyckoff_multiplicity"]
    # Create a list to store the elements with the lowest Wyckoff label
    lowest_wyckoff_elements = []
    # Determine the lowest Wyckoff label between A and B
    min_wyckoff_multiplicity = min(A_lowest_wyckoff_multiplicity, B_lowest_wyckoff_multiplicity)
    # If A or B have the lowest Wyckoff label, add them to the list
    if A_lowest_wyckoff_multiplicity == min_wyckoff_multiplicity:
        lowest_wyckoff_elements.append(A)
    if B_lowest_wyckoff_multiplicity == min_wyckoff_multiplicity:
        lowest_wyckoff_elements.append(B)
    identical_lowest_wyckoff_multiplicity_count = len(lowest_wyckoff_elements)
    data = {
        "WYK_A_lowest_wyckoff": A_lowest_wyckoff_multiplicity,
        "WYK_B_lowest_wyckoff": B_lowest_wyckoff_multiplicity,
        "WYK_identical_lowest_wyckoff_count": identical_lowest_wyckoff_multiplicity_count,
        "WYK_A_sites_total": A_sites_total,
        "WYK_B_sites_total": B_sites_total,
        "WYK_A_multiplicity_total": A_multiplicity_total,
        "WYK_B_multiplicity_total": B_multiplicity_total,
    }
    uni_data = {
        "UNI_WYK_lowest_wyckoff": min_wyckoff_multiplicity,
    }
    return data, uni_data



from cifkit import Cif
from numpy import cos, pi, sqrt


def compute_efficiency(
    cif: Cif,
    radius_dict,
) -> float:
    """Compute the packing efficiency of a crystal structure.

    It uses the number of multiplicities of atoms in the CIF loop values
    and the CIF radii dictionary to calculate the total volume occupied
    by the atoms, and then divides it by the volume of the unit cell.
    """
    elements = cif.unique_elements
    atom_counts = {element: 0 for element in elements}
    atom_site_types = cif._loop_values[1]
    atom_site_symmetry_multiplicities = cif._loop_values[2]
    # Calculate atom counts
    for (
        atom_site_type,
        atom_site_symmetry_multiplicity,
    ) in zip(atom_site_types, atom_site_symmetry_multiplicities):
        if atom_site_type in atom_counts:
            atom_counts[atom_site_type] += int(atom_site_symmetry_multiplicity)
    # Calculate total volume for all atom types
    vol_of_atoms = 0
    for atom, count in atom_counts.items():
        vol_of_atoms += count * (4 / 3) * pi * radius_dict[atom] ** 3
    vol_of_unt_cell = _get_unit_cell_volume(cif.unitcell_lengths, cif.unitcell_angles)
    packing_eff_refined = vol_of_atoms / vol_of_unt_cell
    return float(packing_eff_refined)


def _get_unit_cell_volume(lengths, angles):
    a_length, b_length, c_length = lengths
    alpha_angle, beta_angle, gamma_angle = angles

    volume = (
        a_length
        * b_length
        * c_length
        * sqrt(
            1
            - cos(alpha_angle) ** 2
            - cos(beta_angle) ** 2
            - cos(gamma_angle) ** 2
            + 2 * cos(alpha_angle) * cos(beta_angle) * cos(gamma_angle)
        )
    )
    return volume



import pytest

from SAF.features.coordination.binary import compute_features


def test_features(Th7Rh3_cif, ThRb_elements_tuple):
    result = compute_features(Th7Rh3_cif, ThRb_elements_tuple)
    expected = {
        "CN_AVG_coordination_number": 13.5,
        "CN_AVG_A_atom_count": 10.75,
        "CN_AVG_B_atom_count": 2.75,
        "CN_AVG_polyhedron_volume": 111.40475,
        "CN_AVG_central_atom_to_center_of_mass_dist": 0.15425,
        "CN_AVG_number_of_edges": 34.5,
        "CN_AVG_number_of_faces": 23.0,
        "CN_AVG_shortest_distance_to_face": 2.56825,
        "CN_AVG_shortest_distance_to_edge": 2.61,
        "CN_AVG_volume_of_inscribed_sphere": 75.8885,
        "CN_AVG_packing_efficiency": 0.65125,
        "CN_MIN_coordination_number": 13.5,
        "CN_MIN_A_atom_count": 10.75,
        "CN_MIN_B_atom_count": 2.75,
        "CN_MIN_polyhedron_volume": 111.40475,
        "CN_MIN_central_atom_to_center_of_mass_dist": 0.15425,
        "CN_MIN_number_of_edges": 34.5,
        "CN_MIN_number_of_faces": 23.0,
        "CN_MIN_shortest_distance_to_face": 2.56825,
        "CN_MIN_shortest_distance_to_edge": 2.61,
        "CN_MIN_volume_of_inscribed_sphere": 75.8885,
        "CN_MIN_packing_efficiency": 0.65125,
        "CN_MAX_coordination_number": 13.5,
        "CN_MAX_A_atom_count": 10.75,
        "CN_MAX_B_atom_count": 2.75,
        "CN_MAX_polyhedron_volume": 111.40475,
        "CN_MAX_central_atom_to_center_of_mass_dist": 0.15425,
        "CN_MAX_number_of_edges": 34.5,
        "CN_MAX_number_of_faces": 23.0,
        "CN_MAX_shortest_distance_to_face": 2.56825,
        "CN_MAX_shortest_distance_to_edge": 2.61,
        "CN_MAX_volume_of_inscribed_sphere": 75.8885,
        "CN_MAX_packing_efficiency": 0.65125,
    }

    for key in result:
        assert result[key] == pytest.approx(expected[key], abs=0.01)



from cifkit import Cif

from SAF.features.coordination.helper import get_CN_atom_count_data


def compute_features(cif: Cif, elements):
    avg_CN_metrics, avg_CN_atom_count = get_CN_atom_count_data(cif, elements)
    return {
        "CN_AVG_coordination_number": avg_CN_metrics["avg"]["number_of_vertices"],
        "CN_AVG_A_atom_count": avg_CN_atom_count["avg"]["A_count"],
        "CN_AVG_B_atom_count": avg_CN_atom_count["avg"]["B_count"],
        "CN_AVG_C_atom_count": avg_CN_atom_count["avg"]["C_count"],
        "CN_AVG_D_atom_count": avg_CN_atom_count["avg"]["D_count"],
        "CN_AVG_polyhedron_volume": avg_CN_metrics["avg"]["volume_of_polyhedron"],
        "CN_AVG_central_atom_to_center_of_mass_dist": avg_CN_metrics["avg"]["distance_from_avg_point_to_center"],
        "CN_AVG_number_of_edges": avg_CN_metrics["avg"]["number_of_edges"],
        "CN_AVG_number_of_faces": avg_CN_metrics["avg"]["number_of_faces"],
        "CN_AVG_shortest_distance_to_face": avg_CN_metrics["avg"]["shortest_distance_to_face"],
        "CN_AVG_shortest_distance_to_edge": avg_CN_metrics["avg"]["shortest_distance_to_edge"],
        "CN_AVG_volume_of_inscribed_sphere": avg_CN_metrics["avg"]["volume_of_inscribed_sphere"],
        "CN_AVG_packing_efficiency": avg_CN_metrics["avg"]["packing_efficiency"],
        "CN_MIN_coordination_number": avg_CN_metrics["min"]["number_of_vertices"],
        "CN_MIN_A_atom_count": avg_CN_atom_count["min"]["A_count"],
        "CN_MIN_B_atom_count": avg_CN_atom_count["min"]["B_count"],
        "CN_MIN_C_atom_count": avg_CN_atom_count["min"]["C_count"],
        "CN_MIN_D_atom_count": avg_CN_atom_count["min"]["D_count"],
        "CN_MIN_polyhedron_volume": avg_CN_metrics["min"]["volume_of_polyhedron"],
        "CN_MIN_central_atom_to_center_of_mass_dist": avg_CN_metrics["min"]["distance_from_avg_point_to_center"],
        "CN_MIN_number_of_edges": avg_CN_metrics["min"]["number_of_edges"],
        "CN_MIN_number_of_faces": avg_CN_metrics["min"]["number_of_faces"],
        "CN_MIN_shortest_distance_to_face": avg_CN_metrics["min"]["shortest_distance_to_face"],
        "CN_MIN_shortest_distance_to_edge": avg_CN_metrics["min"]["shortest_distance_to_edge"],
        "CN_MIN_volume_of_inscribed_sphere": avg_CN_metrics["min"]["volume_of_inscribed_sphere"],
        "CN_MIN_packing_efficiency": avg_CN_metrics["min"]["packing_efficiency"],
        "CN_MAX_coordination_number": avg_CN_metrics["max"]["number_of_vertices"],
        "CN_MAX_A_atom_count": avg_CN_atom_count["max"]["A_count"],
        "CN_MAX_B_atom_count": avg_CN_atom_count["max"]["B_count"],
        "CN_MAX_C_atom_count": avg_CN_atom_count["max"]["C_count"],
        "CN_MAX_D_atom_count": avg_CN_atom_count["max"]["D_count"],
        "CN_MAX_polyhedron_volume": avg_CN_metrics["max"]["volume_of_polyhedron"],
        "CN_MAX_central_atom_to_center_of_mass_dist": avg_CN_metrics["max"]["distance_from_avg_point_to_center"],
        "CN_MAX_number_of_edges": avg_CN_metrics["max"]["number_of_edges"],
        "CN_MAX_number_of_faces": avg_CN_metrics["max"]["number_of_faces"],
        "CN_MAX_shortest_distance_to_face": avg_CN_metrics["max"]["shortest_distance_to_face"],
        "CN_MAX_shortest_distance_to_edge": avg_CN_metrics["max"]["shortest_distance_to_edge"],
        "CN_MAX_volume_of_inscribed_sphere": avg_CN_metrics["max"]["volume_of_inscribed_sphere"],
        "CN_MAX_packing_efficiency": avg_CN_metrics["max"]["packing_efficiency"],
    }



import itertools
from typing import Dict, Tuple


def get_min_distances_by_labels(
    shortest_bond_pair_distance: Dict[Tuple[str, str], float],
    elements: list[str],
    labels: list[str] = None,
) -> dict[str, float]:
    """Get bond distances using symbolic labels (like RR, RX, MX) for
    arbitrary elements.

    Parameters
    ----------
        shortest_bond_pair_distance: dict of bond distances keyed by (element1, element2)
        elements: list of actual element symbols
        labels: optional symbolic labels (e.g., ['R', 'M', 'X']) to use for display keys

    Returns
    -------
        Dict[str, float] with symbolic pair keys

    Examples
    --------
    >>> bond_distances = {
    ...     ('Th', 'Th'): 1.1,
    ...     ('Os', 'Os'): 1.2,
    ...     ('Th', 'Os'): 1.3,
    ... }
    >>> get_shortest_bond_distances(bond_distances, ['Th', 'Os'], labels=['A', 'B'])
    {'AA': 1.1, 'BB': 1.2, 'AB': 1.3}
    """
    if labels is None:
        labels = elements

    distances = {}
    for elem1, elem2 in itertools.combinations_with_replacement(elements, 2):
        label1 = labels[elements.index(elem1)]
        label2 = labels[elements.index(elem2)]
        key = f"{label1}{label2}"

        dist = shortest_bond_pair_distance.get((elem1, elem2)) or shortest_bond_pair_distance.get((elem2, elem1))
        if dist is None:
            raise KeyError(f"Missing bond distance for pair ({elem1}, {elem2})")
        distances[key] = dist

    return distances



from deepdiff import DeepDiff

from SAF.features.environment.binary import compute_features


def test_compute_binary_env_features_ThSb(ThSb_cif, ThSb_elements_tuple):
    actual = compute_features(ThSb_cif, ThSb_elements_tuple)
    expected = {
        "ENV_A_shortest_dist_count": 8,
        "ENV_B_shortest_dist_count": 8,
        "ENV_A_avg_shortest_dist_count": 8.0,
        "ENV_B_avg_shortest_dist_count": 8.0,
        "ENV_A_shortest_tol_dist_count": 8,
        "ENV_B_shortest_tol_dist_count": 8,
        "ENV_A_avg_shortest_dist_within_tol_count": 8.0,
        "ENV_B_avg_shortest_dist_within_tol_count": 8.0,
        "ENV_A_second_by_first_shortest_dist": 1.154561824729892,
        "ENV_B_second_by_first_shortest_dist": 1.154561824729892,
        "ENV_A_avg_second_by_first_shortest_dist": 1.154561824729892,
        "ENV_B_avg_second_by_first_shortest_dist": 1.154561824729892,
        "ENV_A_second_shortest_dist_count": 6,
        "ENV_B_second_shortest_dist_count": 6,
        "ENV_A_avg_second_shortest_dist_count": 6.0,
        "ENV_B_avg_second_shortest_dist_count": 6.0,
        "ENV_A_homoatomic_dist_by_shortest_dist": 1.154561824729892,
        "ENV_B_homoatomic_dist_by_shortest_dist": 1.154561824729892,
        "ENV_A_avg_homoatomic_dist_by_shortest_dist": 1.154561824729892,
        "ENV_B_avg_homoatomic_dist_by_shortest_dist": 1.154561824729892,
        "ENV_A_count_at_A_shortest_dist": 0,
        "ENV_A_count_at_B_shortest_dist": 8,
        "ENV_A_avg_count_at_A_shortest_dist": 0.0,
        "ENV_A_avg_count_at_B_shortest_dist": 8.0,
        "ENV_B_count_at_A_shortest_dist": 8,
        "ENV_B_count_at_B_shortest_dist": 0,
        "ENV_B_avg_count_at_A_shortest_dist": 8.0,
        "ENV_B_avg_count_at_B_shortest_dist": 0.0,
    }
    diff = DeepDiff(actual, expected, significant_digits=3)
    assert diff == {}


def test_compute_binary_env_features_RhSb2(RhSb2_cif, RhSb_elements_tuple):
    actual = compute_features(RhSb2_cif, RhSb_elements_tuple)
    expected = {
        "ENV_A_shortest_dist_count": 1,
        "ENV_B_shortest_dist_count": 1,
        "ENV_A_avg_shortest_dist_count": 1.0,
        "ENV_B_avg_shortest_dist_count": 1.0,
        "ENV_A_shortest_tol_dist_count": 3,
        "ENV_B_shortest_tol_dist_count": 3,
        "ENV_A_avg_shortest_dist_within_tol_count": 3.0,
        "ENV_B_avg_shortest_dist_within_tol_count": 3.0,
        "ENV_A_second_by_first_shortest_dist": 1.0151033386327504,
        "ENV_B_second_by_first_shortest_dist": 1.0151033386327504,
        "ENV_A_avg_second_by_first_shortest_dist": 1.0151033386327504,
        "ENV_B_avg_second_by_first_shortest_dist": 1.0084943540825893,
        "ENV_A_second_shortest_dist_count": 1,
        "ENV_B_second_shortest_dist_count": 1,
        "ENV_A_avg_second_shortest_dist_count": 1.0,
        "ENV_B_avg_second_shortest_dist_count": 1.0,
        "ENV_A_homoatomic_dist_by_shortest_dist": 1.1971383147853736,
        "ENV_B_homoatomic_dist_by_shortest_dist": 1.1244038155802862,
        "ENV_A_avg_homoatomic_dist_by_shortest_dist": 1.1971383147853736,
        "ENV_B_avg_homoatomic_dist_by_shortest_dist": 1.095572948514125,
        "ENV_A_count_at_A_shortest_dist": 0,
        "ENV_A_count_at_B_shortest_dist": 1,
        "ENV_A_avg_count_at_A_shortest_dist": 0.0,
        "ENV_A_avg_count_at_B_shortest_dist": 1.0,
        "ENV_B_count_at_A_shortest_dist": 1,
        "ENV_B_count_at_B_shortest_dist": 0,
        "ENV_B_avg_count_at_A_shortest_dist": 1.0,
        "ENV_B_avg_count_at_B_shortest_dist": 0.0,
    }
    diff = DeepDiff(actual, expected, significant_digits=3)
    assert diff == {}



import numpy as np

from SAF.features.environment.util import count_first_second_min_dist, extract_best_labels


def get_nearest_connections_no_coords(connections, count=15):
    """Return the first `count` connections for each site label,
    including only the element label and distance.

    This simplifies inspection/debugging.
    """
    return {k: [(elem, dist) for (elem, dist, *_) in v[:count]] for k, v in list(connections.items())[:count]}


def test_count_first_second_min_dist_binary(ThSb_cif):
    connections = ThSb_cif.connections
    first_second_min_dists = get_nearest_connections_no_coords(connections)
    assert first_second_min_dists == {
        "Sb1": [
            ("Th1", 3.332),
            ("Th1", 3.332),
            ("Th1", 3.332),
            ("Th1", 3.332),
            ("Th1", 3.332),
            ("Th1", 3.332),
            ("Th1", 3.332),
            ("Th1", 3.332),
            ("Sb1", 3.847),
            ("Sb1", 3.847),
            ("Sb1", 3.847),
            ("Sb1", 3.847),
            ("Sb1", 3.847),
            ("Sb1", 3.847),
            ("Sb1", 5.44),
        ],
        "Th1": [
            ("Sb1", 3.332),
            ("Sb1", 3.332),
            ("Sb1", 3.332),
            ("Sb1", 3.332),
            ("Sb1", 3.332),
            ("Sb1", 3.332),
            ("Sb1", 3.332),
            ("Sb1", 3.332),
            ("Th1", 3.847),
            ("Th1", 3.847),
            ("Th1", 3.847),
            ("Th1", 3.847),
            ("Th1", 3.847),
            ("Th1", 3.847),
            ("Th1", 5.44),
        ],
    }
    first_second_min_dists = count_first_second_min_dist(connections)
    expected = {
        "Sb1": {
            "shortest_dist": np.float64(3.332),
            "second_shortest_dist": np.float64(3.847),
            "counts": {np.float64(3.332): 8, np.float64(3.847): 6},
        },
        "Th1": {
            "shortest_dist": np.float64(3.332),
            "second_shortest_dist": np.float64(3.847),
            "counts": {np.float64(3.332): 8, np.float64(3.847): 6},
        },
    }
    assert first_second_min_dists == expected


def test_count_first_second_min_dist_ternary(URhIn_cif):
    connections = URhIn_cif.connections
    first_second_min_dists = get_nearest_connections_no_coords(connections)
    assert first_second_min_dists == {
        "In1": [
            ("Rh2", 2.697),
            ("Rh2", 2.697),
            ("Rh1", 2.852),
            ("Rh1", 2.852),
            ("U1", 3.21),
            ("U1", 3.21),
            ("In1", 3.244),
            ("In1", 3.244),
            ("U1", 3.294),
            ("U1", 3.294),
            ("U1", 3.294),
            ("U1", 3.294),
            ("In1", 3.881),
            ("In1", 3.881),
            ("Rh1", 4.705),
        ],
        "U1": [
            ("Rh1", 2.983),
            ("Rh1", 2.983),
            ("Rh1", 2.983),
            ("Rh1", 2.984),
            ("Rh2", 3.046),
            ("In1", 3.21),
            ("In1", 3.21),
            ("In1", 3.294),
            ("In1", 3.294),
            ("In1", 3.294),
            ("In1", 3.294),
            ("U1", 3.881),
            ("U1", 3.881),
            ("U1", 3.925),
            ("U1", 3.925),
        ],
        "Rh1": [
            ("In1", 2.852),
            ("In1", 2.852),
            ("In1", 2.853),
            ("U1", 2.983),
            ("U1", 2.983),
            ("U1", 2.983),
            ("U1", 2.984),
            ("U1", 2.984),
            ("U1", 2.984),
            ("Rh1", 3.881),
            ("Rh1", 3.881),
            ("Rh1", 4.316),
            ("Rh1", 4.316),
            ("Rh1", 4.316),
            ("In1", 4.705),
        ],
        "Rh2": [
            ("In1", 2.697),
            ("In1", 2.697),
            ("In1", 2.697),
            ("In1", 2.697),
            ("In1", 2.697),
            ("In1", 2.697),
            ("U1", 3.046),
            ("U1", 3.046),
            ("U1", 3.046),
            ("Rh2", 3.881),
            ("Rh2", 3.881),
            ("U1", 4.43),
            ("U1", 4.43),
            ("U1", 4.43),
            ("Rh1", 4.732),
        ],
    }
    first_second_min_dists = count_first_second_min_dist(connections)
    assert first_second_min_dists == {
        "In1": {"shortest_dist": 2.697, "second_shortest_dist": 2.852, "counts": {2.697: 2, 2.852: 2}},
        "U1": {"shortest_dist": 2.983, "second_shortest_dist": 2.984, "counts": {2.983: 3, 2.984: 1}},
        "Rh1": {"shortest_dist": 2.852, "second_shortest_dist": 2.853, "counts": {2.852: 2, 2.853: 1}},
        "Rh2": {"shortest_dist": 2.697, "second_shortest_dist": 3.046, "counts": {2.697: 6, 3.046: 3}},
    }


def test_extract_best_labels_binary(URhIn_cif):
    connections = URhIn_cif.connections
    first_second_min_dists = count_first_second_min_dist(connections)
    best_labels = extract_best_labels(first_second_min_dists)
    expected = {
        "In": {
            "label_count": 1,
            "shortest_dist_total_count": 2,
            "second_shortest_dist_count": 2,
            "avg_shortest_dist": 2.0,
            "avg_second_shortest_dist_count": 2.0,
            "best_label": "In1",
            "best_label_details": {"shortest_dist": 2.697, "second_shortest_dist": 2.852, "counts": {2.697: 2, 2.852: 2}},
        },
        "U": {
            "label_count": 1,
            "shortest_dist_total_count": 3,
            "second_shortest_dist_count": 1,
            "avg_shortest_dist": 3.0,
            "avg_second_shortest_dist_count": 1.0,
            "best_label": "U1",
            "best_label_details": {"shortest_dist": 2.983, "second_shortest_dist": 2.984, "counts": {2.983: 3, 2.984: 1}},
        },
        "Rh": {
            "label_count": 2,
            "shortest_dist_total_count": 8,  # correct Rh1 (2) + Rh2 (6)
            "second_shortest_dist_count": 4,  # correct Rh1 (1) + Rh2 (3)
            "avg_shortest_dist": 4.0,  # correct 8 / 2
            "avg_second_shortest_dist_count": 2.0,  # correct 4 / 2
            "best_label": "Rh2",  # correct Rh2 has the shortest distance of 2.697
            "best_label_details": {"shortest_dist": 2.697, "second_shortest_dist": 3.046, "counts": {2.697: 6, 3.046: 3}},
        },
    }
    assert best_labels == expected



from SAF.features.wyc.quaternary import compute_features


def test_compute_features_Th4RhInGe4(Tb4RhInGe4_cif, TbRhInGe_elements_tuple):
    # Let A = Tb, B = Rh, C = In, D = Ge
    wyk_actual, wyk_uni_actual = compute_features(Tb4RhInGe4_cif, TbRhInGe_elements_tuple)
    expected = {
        "WYK_A_lowest_wyckoff": 4,
        "WYK_B_lowest_wyckoff": 4,
        "WYK_C_lowest_wyckoff": 2,
        "WYK_D_lowest_wyckoff": 4,
        "WYK_identical_lowest_wyckoff_count": 1,
        "WYK_A_sites_total": 4,
        "WYK_B_sites_total": 1,
        "WYK_C_sites_total": 2,
        "WYK_D_sites_total": 4,
        "WYK_A_multiplicity_total": 16,
        "WYK_B_multiplicity_total": 4,
        "WYK_C_multiplicity_total": 4,
        "WYK_D_multiplicity_total": 16,
    }
    uni_expected = {"UNI_WYK_lowest_wyckoff": 2}

    print("WYK Actual:", wyk_actual)
    print("WYK Uni Actual:", wyk_uni_actual)
    assert wyk_actual == expected
    assert wyk_uni_actual == uni_expected



import pytest

from SAF.features.environment.util import (
    count_first_second_min_dist,
    extract_avg_shortest_dist_with_tol,
    extract_best_labels,
    extract_shortest_dist_with_tol,
    get_avg_second_by_first_shortest_dist_ratio,
)


@pytest.fixture(scope="function")
def Th7Rh3_site_dist_info():
    return {
        "Rh1": {
            "shortest_dist": 2.869,
            "second_shortest_dist": 2.95,
            "counts": {2.869: 2, 2.95: 1},
        },
        "Th1": {
            "shortest_dist": 2.965,
            "second_shortest_dist": 3.496,
            "counts": {2.965: 2, 3.496: 2},
        },
        "Th2": {
            "shortest_dist": 2.869,
            "second_shortest_dist": 2.95,
            "counts": {2.869: 2, 2.95: 1},
        },
        "Th3": {
            "shortest_dist": 3.019,
            "second_shortest_dist": 3.643,
            "counts": {3.019: 3, 3.643: 3},
        },
    }


@pytest.fixture(scope="function")
def binary_ThSb_cif_site_dist_info():
    return {
        "Sb1": {
            "shortest_dist": 3.332,
            "second_shortest_dist": 3.847,
            "counts": {3.332: 8, 3.847: 6},
        },
        "Th1": {
            "shortest_dist": 3.332,
            "second_shortest_dist": 3.847,
            "counts": {3.332: 8, 3.847: 6},
        },
    }


def test_find_shortest_distances_binary_ThSb(ThSb_cif, binary_ThSb_cif_site_dist_info):
    result = count_first_second_min_dist(ThSb_cif.connections)
    assert result == binary_ThSb_cif_site_dist_info


def test_find_shortest_distances_binary_Th7Rh3(Th7Rh3_cif, Th7Rh3_site_dist_info):
    result = count_first_second_min_dist(Th7Rh3_cif.connections)
    assert result == Th7Rh3_site_dist_info


def test_extract_best_labels_Th7Rh3(Th7Rh3_site_dist_info):
    result = extract_best_labels(Th7Rh3_site_dist_info)

    assert result == {
        "Rh": {
            "label_count": 1,
            "shortest_dist_total_count": 2,
            "second_shortest_dist_count": 1,
            "avg_shortest_dist": 2.0,
            "avg_second_shortest_dist_count": 1.0,
            "best_label": "Rh1",
            "best_label_details": {
                "shortest_dist": 2.869,
                "second_shortest_dist": 2.95,
                "counts": {2.869: 2, 2.95: 1},
            },
        },
        "Th": {
            "label_count": 3,
            "shortest_dist_total_count": 7,
            "second_shortest_dist_count": 6,
            "avg_shortest_dist": 2.3333333333333335,
            "avg_second_shortest_dist_count": 2.0,
            "best_label": "Th2",
            "best_label_details": {
                "shortest_dist": 2.869,
                "second_shortest_dist": 2.95,
                "counts": {2.869: 2, 2.95: 1},
            },
        },
    }


def test_extract_best_labels_531015(
    binary_ThSb_cif_site_dist_info,
):
    result = extract_best_labels(binary_ThSb_cif_site_dist_info)
    assert result == {
        "Sb": {
            "label_count": 1,
            "shortest_dist_total_count": 8,
            "second_shortest_dist_count": 6,
            "avg_shortest_dist": 8.0,
            "avg_second_shortest_dist_count": 6.0,
            "best_label": "Sb1",
            "best_label_details": {
                "shortest_dist": 3.332,
                "second_shortest_dist": 3.847,
                "counts": {3.332: 8, 3.847: 6},
            },
        },
        "Th": {
            "label_count": 1,
            "shortest_dist_total_count": 8,
            "second_shortest_dist_count": 6,
            "avg_shortest_dist": 8.0,
            "avg_second_shortest_dist_count": 6.0,
            "best_label": "Th1",
            "best_label_details": {
                "shortest_dist": 3.332,
                "second_shortest_dist": 3.847,
                "counts": {3.332: 8, 3.847: 6},
            },
        },
    }


def test_extract_shortest_dist_with_tol(ThSb_cif, binary_ThSb_cif_site_dist_info):
    best_label_dist_info = extract_best_labels(binary_ThSb_cif_site_dist_info)
    result = extract_shortest_dist_with_tol(best_label_dist_info, ThSb_cif.connections)

    assert result == {
        "Sb": {"shortest_dist_count_within_tol": 8},
        "Th": {"shortest_dist_count_within_tol": 8},
    }


def test_extract_avg_shortest_dist_with_tol_Th7Rh3(Th7Rh3_cif):
    result = extract_avg_shortest_dist_with_tol(Th7Rh3_cif.connections)

    assert result == {
        "Rh": {
            "total_shortest_dist_within_tol_count": 5,
            "avg_shortest_dist_within_tol_count": 5.0,
        },
        "Th": {
            "total_shortest_dist_within_tol_count": 8,
            "avg_shortest_dist_within_tol_count": 2.6666666666666665,
        },
    }


def test_extract_avg_shortest_dist_with_tol(ThSb_cif):
    result = extract_avg_shortest_dist_with_tol(ThSb_cif.connections)
    assert result == {
        "Sb": {
            "total_shortest_dist_within_tol_count": 8,
            "avg_shortest_dist_within_tol_count": 8.0,
        },
        "Th": {
            "total_shortest_dist_within_tol_count": 8,
            "avg_shortest_dist_within_tol_count": 8.0,
        },
    }


def test_get_first_by_second_shortest_dist_ratio_ThSb(binary_ThSb_cif_site_dist_info, ThSb_cif):
    result = get_avg_second_by_first_shortest_dist_ratio(binary_ThSb_cif_site_dist_info, ThSb_cif.connections)

    assert result == {
        "Sb": {"avg_second_by_first_shortest_dist": 1.154561824729892},
        "Th": {"avg_second_by_first_shortest_dist": 1.154561824729892},
    }


def test_get_first_by_second_shortest_dist_ratio_Th7Rh3(Th7Rh3_site_dist_info, Th7Rh3_cif):
    result = get_avg_second_by_first_shortest_dist_ratio(Th7Rh3_site_dist_info, Th7Rh3_cif.connections)

    assert result == {
        "Rh": {"avg_second_by_first_shortest_dist": 1.028232833739979},
        "Th": {"avg_second_by_first_shortest_dist": 1.1380043890215203},
    }



"""Unit tests for __version__.py."""

import SAF  # noqa


def test_package_version():
    """Ensure the package version is defined and not set to the initial
    placeholder."""
    assert hasattr(SAF, "__version__")
    assert SAF.__version__ != "0.0.0"



from cifkit import Cif

from SAF.features.environment.binary_helper import (
    compute_avg_homoatomic_dist_by_site_shortest_dist,
    compute_homoatomic_dist_by_site_shortest_dist,
    get_A_and_B_count_in_best_label_per_element,
    get_avg_A_and_B_count_in_per_element,
)
from SAF.features.environment.util import (
    count_first_second_min_dist,
    extract_avg_shortest_dist_with_tol,
    extract_best_labels,
    extract_shortest_dist_with_tol,
    get_avg_second_by_first_shortest_dist_ratio,
)


def compute_features(cif: Cif, elements):
    A, B = elements
    connections = cif.connections
    (
        A_avg_homoatomic_dist_by_shortest_dist,
        B_avg_homoatomic_dist_by_shortest_dist,
    ) = compute_avg_homoatomic_dist_by_site_shortest_dist(connections, A, B)
    (
        A_count_at_A_shortest_dist,
        A_count_at_B_shortest_dist,
        B_count_at_A_shortest_dist,
        B_count_at_B_shortest_dist,
    ) = get_A_and_B_count_in_best_label_per_element(connections, A, B)
    (
        A_avg_count_at_A_shortest_dist,
        A_avg_count_at_B_shortest_dist,
        B_avg_count_at_A_shortest_dist,
        B_avg_count_at_B_shortest_dist,
    ) = get_avg_A_and_B_count_in_per_element(connections, A, B)
    # Compute the first, second shortest distances and count
    first_second_dist_per_label_data = count_first_second_min_dist(connections)
    # Compute the best site for each label, determined by the shortest distance
    best_site_data = extract_best_labels(first_second_dist_per_label_data)
    # Tol result
    tol_results = extract_shortest_dist_with_tol(best_site_data, connections)
    A_shortest_dist_count_within_tol = tol_results[A]["shortest_dist_count_within_tol"]
    B_shortest_dist_count_within_tol = tol_results[B]["shortest_dist_count_within_tol"]
    avg_tol_results = extract_avg_shortest_dist_with_tol(connections)
    A_avg_shortest_dist_within_tol_count = avg_tol_results[A]["avg_shortest_dist_within_tol_count"]
    B_avg_shortest_dist_within_tol_count = avg_tol_results[B]["avg_shortest_dist_within_tol_count"]
    A_best_label = best_site_data[A]["best_label"]
    B_best_label = best_site_data[B]["best_label"]

    (
        A_homoatomic_dist_by_shortest_dist,
        B_homoatomic_dist_by_shortest_dist,
    ) = compute_homoatomic_dist_by_site_shortest_dist(connections, A_best_label, B_best_label)
    # First shortest distance
    A_shortest_dist = best_site_data[A]["best_label_details"]["shortest_dist"]
    B_shortest_dist = best_site_data[B]["best_label_details"]["shortest_dist"]
    # Second shortest distance
    A_second_shortest_dist = best_site_data[A]["best_label_details"]["second_shortest_dist"]
    B_second_shortest_dist = best_site_data[B]["best_label_details"]["second_shortest_dist"]
    # Shorest distance count from best site
    A_shortest_dist_count = best_site_data[A]["best_label_details"]["counts"][A_shortest_dist]
    B_shortest_dist_count = best_site_data[B]["best_label_details"]["counts"][B_shortest_dist]
    # Second shorest distance count from best site
    A_second_shortest_dist_count = best_site_data[A]["best_label_details"]["counts"][A_second_shortest_dist]
    B_second_shortest_dist_count = best_site_data[B]["best_label_details"]["counts"][B_second_shortest_dist]
    # Avg shortest distance count across site labels per element
    A_avg_shortest_dist_count = best_site_data[A]["avg_shortest_dist"]
    B_avg_shortest_dist_count = best_site_data[B]["avg_shortest_dist"]
    # Avg second shortest distance count across site labels per element
    A_avg_second_shortest_dist_count = best_site_data[A]["avg_second_shortest_dist_count"]
    B_avg_second_shortest_dist_count = best_site_data[B]["avg_second_shortest_dist_count"]
    # Get avg second by first shortest distance ratio across site labels per element
    avg_second_by_first_dist = get_avg_second_by_first_shortest_dist_ratio(first_second_dist_per_label_data, connections)
    A_avg_second_by_first_shortest_dist = avg_second_by_first_dist[A]["avg_second_by_first_shortest_dist"]
    B_avg_second_by_first_shortest_dist = avg_second_by_first_dist[B]["avg_second_by_first_shortest_dist"]
    data = {
        "ENV_A_shortest_dist_count": A_shortest_dist_count,
        "ENV_B_shortest_dist_count": B_shortest_dist_count,
        "ENV_A_avg_shortest_dist_count": A_avg_shortest_dist_count,
        "ENV_B_avg_shortest_dist_count": B_avg_shortest_dist_count,
        "ENV_A_shortest_tol_dist_count": A_shortest_dist_count_within_tol,
        "ENV_B_shortest_tol_dist_count": B_shortest_dist_count_within_tol,
        "ENV_A_avg_shortest_dist_within_tol_count": A_avg_shortest_dist_within_tol_count,
        "ENV_B_avg_shortest_dist_within_tol_count": B_avg_shortest_dist_within_tol_count,
        "ENV_A_second_by_first_shortest_dist": A_second_shortest_dist / A_shortest_dist,
        "ENV_B_second_by_first_shortest_dist": B_second_shortest_dist / B_shortest_dist,
        "ENV_A_avg_second_by_first_shortest_dist": A_avg_second_by_first_shortest_dist,
        "ENV_B_avg_second_by_first_shortest_dist": B_avg_second_by_first_shortest_dist,
        "ENV_A_second_shortest_dist_count": A_second_shortest_dist_count,
        "ENV_B_second_shortest_dist_count": B_second_shortest_dist_count,
        "ENV_A_avg_second_shortest_dist_count": A_avg_second_shortest_dist_count,
        "ENV_B_avg_second_shortest_dist_count": B_avg_second_shortest_dist_count,
        "ENV_A_homoatomic_dist_by_shortest_dist": A_homoatomic_dist_by_shortest_dist,
        "ENV_B_homoatomic_dist_by_shortest_dist": B_homoatomic_dist_by_shortest_dist,
        "ENV_A_avg_homoatomic_dist_by_shortest_dist": A_avg_homoatomic_dist_by_shortest_dist,
        "ENV_B_avg_homoatomic_dist_by_shortest_dist": B_avg_homoatomic_dist_by_shortest_dist,
        "ENV_A_count_at_A_shortest_dist": A_count_at_A_shortest_dist,
        "ENV_A_count_at_B_shortest_dist": A_count_at_B_shortest_dist,
        "ENV_A_avg_count_at_A_shortest_dist": A_avg_count_at_A_shortest_dist,
        "ENV_A_avg_count_at_B_shortest_dist": A_avg_count_at_B_shortest_dist,
        "ENV_B_count_at_A_shortest_dist": B_count_at_A_shortest_dist,
        "ENV_B_count_at_B_shortest_dist": B_count_at_B_shortest_dist,
        "ENV_B_avg_count_at_A_shortest_dist": B_avg_count_at_A_shortest_dist,
        "ENV_B_avg_count_at_B_shortest_dist": B_avg_count_at_B_shortest_dist,
    }
    return data



from cifkit import Cif

from SAF.features.environment.quaternary_helper import (
    compute_avg_homoatomic_dist_by_site_shortest_dist,
    compute_homoatomic_dist_by_site_shortest_dist,
    get_A_and_B_and_C_and_D_count_in_best_label_per_element,
    get_avg_A_and_B_and_C_and_D_count_in_per_element,
)
from SAF.features.environment.util import (
    count_first_second_min_dist,
    extract_avg_shortest_dist_with_tol,
    extract_best_labels,
    extract_shortest_dist_with_tol,
    get_avg_second_by_first_shortest_dist_ratio,
)


def compute_features(cif: Cif, elements):
    connections = cif.connections
    A, B, C, D = elements

    (
        A_avg_homoatomic_dist_by_shortest_dist,
        B_avg_homoatomic_dist_by_shortest_dist,
        C_avg_homoatomic_dist_by_shortest_dist,
        D_avg_homoatomic_dist_by_shortest_dist,
    ) = compute_avg_homoatomic_dist_by_site_shortest_dist(connections, A, B, C, D)
    (
        A_count_at_A_shortest_dist,
        A_count_at_B_shortest_dist,
        A_count_at_C_shortest_dist,
        A_count_at_D_shortest_dist,
        B_count_at_A_shortest_dist,
        B_count_at_B_shortest_dist,
        B_count_at_C_shortest_dist,
        B_count_at_D_shortest_dist,
        C_count_at_A_shortest_dist,
        C_count_at_B_shortest_dist,
        C_count_at_C_shortest_dist,
        C_count_at_D_shortest_dist,
        D_count_at_A_shortest_dist,
        D_count_at_B_shortest_dist,
        D_count_at_C_shortest_dist,
        D_count_at_D_shortest_dist,
    ) = get_A_and_B_and_C_and_D_count_in_best_label_per_element(connections, A, B, C, D)
    (
        A_avg_count_at_A_shorest_dist,
        A_avg_count_at_B_shorest_dist,
        A_avg_count_at_C_shorest_dist,
        A_avg_count_at_D_shorest_dist,
        B_avg_count_at_A_shorest_dist,
        B_avg_count_at_B_shorest_dist,
        B_avg_count_at_C_shorest_dist,
        B_avg_count_at_D_shorest_dist,
        C_avg_count_at_A_shorest_dist,
        C_avg_count_at_B_shorest_dist,
        C_avg_count_at_C_shorest_dist,
        C_avg_count_at_D_shorest_dist,
        D_avg_count_at_A_shorest_dist,
        D_avg_count_at_B_shorest_dist,
        D_avg_count_at_C_shorest_dist,
        D_avg_count_at_D_shorest_dist,
    ) = get_avg_A_and_B_and_C_and_D_count_in_per_element(connections, A, B, C, D)
    # 1. Compute the first, second shortest distances and count
    first_second_dist_per_site_data = count_first_second_min_dist(connections)
    # 2. Compute the best site for each label, determined by the shortest distance
    best_site_data = extract_best_labels(first_second_dist_per_site_data)
    # 3. Tol result
    tol_results = extract_shortest_dist_with_tol(best_site_data, connections)
    A_shortest_dist_count_within_tol = tol_results[A]["shortest_dist_count_within_tol"]
    B_shortest_dist_count_within_tol = tol_results[B]["shortest_dist_count_within_tol"]
    C_shortest_dist_count_within_tol = tol_results[C]["shortest_dist_count_within_tol"]
    D_shortest_dist_count_within_tol = tol_results[D]["shortest_dist_count_within_tol"]
    avg_tol_results = extract_avg_shortest_dist_with_tol(connections)
    A_avg_shortest_dist_within_tol_count = avg_tol_results[A]["avg_shortest_dist_within_tol_count"]
    B_avg_shortest_dist_within_tol_count = avg_tol_results[B]["avg_shortest_dist_within_tol_count"]
    C_avg_shortest_dist_within_tol_count = avg_tol_results[C]["avg_shortest_dist_within_tol_count"]
    D_avg_shortest_dist_within_tol_count = avg_tol_results[D]["avg_shortest_dist_within_tol_count"]
    A_best_label = best_site_data[A]["best_label"]
    B_best_label = best_site_data[B]["best_label"]
    C_best_label = best_site_data[C]["best_label"]
    D_best_label = best_site_data[D]["best_label"]
    (
        A_homoatomic_dist_by_shortest_dist,
        B_homoatomic_dist_by_shortest_dist,
        C_homoatomic_dist_by_shortest_dist,
        D_homoatomic_dist_by_shortest_dist,
    ) = compute_homoatomic_dist_by_site_shortest_dist(connections, A_best_label, B_best_label, C_best_label, D_best_label)
    # First shortest distance
    A_shortest_dist = best_site_data[A]["best_label_details"]["shortest_dist"]
    B_shortest_dist = best_site_data[B]["best_label_details"]["shortest_dist"]
    C_shortest_dist = best_site_data[C]["best_label_details"]["shortest_dist"]
    D_shortest_dist = best_site_data[D]["best_label_details"]["shortest_dist"]
    # Second shortest distance
    A_second_shortest_dist = best_site_data[A]["best_label_details"]["second_shortest_dist"]
    B_second_shortest_dist = best_site_data[B]["best_label_details"]["second_shortest_dist"]
    C_second_shortest_dist = best_site_data[C]["best_label_details"]["second_shortest_dist"]
    D_second_shortest_dist = best_site_data[D]["best_label_details"]["second_shortest_dist"]
    # First shortest distance count
    A_shortest_dist_count = best_site_data[A]["best_label_details"]["counts"][A_shortest_dist]
    B_shortest_dist_count = best_site_data[B]["best_label_details"]["counts"][B_shortest_dist]
    C_shortest_dist_count = best_site_data[C]["best_label_details"]["counts"][C_shortest_dist]
    D_shortest_dist_count = best_site_data[D]["best_label_details"]["counts"][D_shortest_dist]
    # Second shortest distance count
    A_second_shortest_dist_count = best_site_data[A]["best_label_details"]["counts"][A_second_shortest_dist]
    B_second_shortest_dist_count = best_site_data[B]["best_label_details"]["counts"][B_second_shortest_dist]
    C_second_shortest_dist_count = best_site_data[C]["best_label_details"]["counts"][C_second_shortest_dist]
    D_second_shortest_dist_count = best_site_data[D]["best_label_details"]["counts"][D_second_shortest_dist]
    # Avg shortest distance count across site labels per element
    A_avg_shortest_dist_count = best_site_data[A]["avg_shortest_dist"]
    B_avg_shortest_dist_count = best_site_data[B]["avg_shortest_dist"]
    C_avg_shortest_dist_count = best_site_data[C]["avg_shortest_dist"]
    D_avg_shortest_dist_count = best_site_data[D]["avg_shortest_dist"]
    # Avg second shortest distance count across site labels per element
    A_avg_second_shortest_dist_count = best_site_data[A]["avg_second_shortest_dist_count"]
    B_avg_second_shortest_dist_count = best_site_data[B]["avg_second_shortest_dist_count"]
    C_avg_second_shortest_dist_count = best_site_data[C]["avg_second_shortest_dist_count"]
    D_avg_second_shortest_dist_count = best_site_data[D]["avg_second_shortest_dist_count"]
    # Get avg second by first shortest distance ratio across site labels per element
    avg_second_by_first_dist = get_avg_second_by_first_shortest_dist_ratio(first_second_dist_per_site_data, connections)
    A_avg_second_by_first_shortest_dist = avg_second_by_first_dist[A]["avg_second_by_first_shortest_dist"]
    B_avg_second_by_first_shortest_dist = avg_second_by_first_dist[B]["avg_second_by_first_shortest_dist"]
    C_avg_second_by_first_shortest_dist = avg_second_by_first_dist[C]["avg_second_by_first_shortest_dist"]
    D_avg_second_by_first_shortest_dist = avg_second_by_first_dist[D]["avg_second_by_first_shortest_dist"]

    results = {
        # Shortest distance count
        "ENV_A_shortest_dist_count": A_shortest_dist_count,
        "ENV_B_shortest_dist_count": B_shortest_dist_count,
        "ENV_C_shortest_dist_count": C_shortest_dist_count,
        "ENV_D_shortest_dist_count": D_shortest_dist_count,
        # Avg shortest distance count
        "ENV_A_avg_shortest_dist_count": A_avg_shortest_dist_count,
        "ENV_B_avg_shortest_dist_count": B_avg_shortest_dist_count,
        "ENV_C_avg_shortest_dist_count": C_avg_shortest_dist_count,
        "ENV_D_avg_shortest_dist_count": D_avg_shortest_dist_count,
        # Shortest distance count within tolerance
        "ENV_A_shortest_tol_dist_count": A_shortest_dist_count_within_tol,
        "ENV_B_shortest_tol_dist_count": B_shortest_dist_count_within_tol,
        "ENV_C_shortest_tol_dist_count": C_shortest_dist_count_within_tol,
        "ENV_D_shortest_tol_dist_count": D_shortest_dist_count_within_tol,
        # Avg shortest distance count within tolerance
        "ENV_A_avg_shortest_dist_within_tol_count": A_avg_shortest_dist_within_tol_count,
        "ENV_B_avg_shortest_dist_within_tol_count": B_avg_shortest_dist_within_tol_count,
        "ENV_C_avg_shortest_dist_within_tol_count": C_avg_shortest_dist_within_tol_count,
        "ENV_D_avg_shortest_dist_within_tol_count": D_avg_shortest_dist_within_tol_count,
        # Second by first shortest distance ratio
        "ENV_A_second_by_first_shortest_dist": A_second_shortest_dist / A_shortest_dist,
        "ENV_B_second_by_first_shortest_dist": B_second_shortest_dist / B_shortest_dist,
        "ENV_C_second_by_first_shortest_dist": C_second_shortest_dist / C_shortest_dist,
        "ENV_D_second_by_first_shortest_dist": D_second_shortest_dist / D_shortest_dist,
        # Avg second by first shortest distance ratio
        "ENV_A_avg_second_by_first_shortest_dist": A_avg_second_by_first_shortest_dist,
        "ENV_B_avg_second_by_first_shortest_dist": B_avg_second_by_first_shortest_dist,
        "ENV_C_avg_second_by_first_shortest_dist": C_avg_second_by_first_shortest_dist,
        "ENV_D_avg_second_by_first_shortest_dist": D_avg_second_by_first_shortest_dist,
        # Second shortest distance count
        "ENV_A_second_shortest_dist_count": A_second_shortest_dist_count,
        "EnV_B_second_shortest_dist_count": B_second_shortest_dist_count,
        "ENV_C_second_shortest_dist_count": C_second_shortest_dist_count,
        "ENV_D_second_shortest_dist_count": D_second_shortest_dist_count,
        # Avg second shortest distance count
        "ENV_A_avg_second_shortest_dist_count": A_avg_second_shortest_dist_count,
        "ENV_B_avg_second_shortest_dist_count": B_avg_second_shortest_dist_count,
        "ENV_C_avg_second_shortest_dist_count": C_avg_second_shortest_dist_count,
        "ENV_D_avg_second_shortest_dist_count": D_avg_second_shortest_dist_count,
        # Homoatomic distance by shortest distance
        "ENV_A_homoatomic_dist_by_shortest_dist": A_homoatomic_dist_by_shortest_dist,
        "ENV_B_homoatomic_dist_by_shortest_dist": B_homoatomic_dist_by_shortest_dist,
        "ENV_C_homoatomic_dist_by_shortest_dist": C_homoatomic_dist_by_shortest_dist,
        "ENV_D_homoatomic_dist_by_shortest_dist": D_homoatomic_dist_by_shortest_dist,
        # Avg homoatomic distance by shortest distance
        "ENV_A_avg_homoatomic_dist_by_shortest_dist": A_avg_homoatomic_dist_by_shortest_dist,
        "ENV_B_avg_homoatomic_dist_by_shortest_dist": B_avg_homoatomic_dist_by_shortest_dist,
        "ENV_C_avg_homoatomic_dist_by_shortest_dist": C_avg_homoatomic_dist_by_shortest_dist,
        "ENV_D_avg_homoatomic_dist_by_shortest_dist": D_avg_homoatomic_dist_by_shortest_dist,
        # Count at shortest distance
        "ENV_A_count_at_A_shortest_dist": A_count_at_A_shortest_dist,
        "ENV_B_count_at_A_shortest_dist": B_count_at_A_shortest_dist,
        "ENV_C_count_at_A_shortest_dist": C_count_at_A_shortest_dist,
        "ENV_D_count_at_A_shortest_dist": D_count_at_A_shortest_dist,
        "ENV_A_avg_count_at_A_shortest_dist": A_avg_count_at_A_shorest_dist,
        "ENV_B_avg_count_at_A_shortest_dist": B_avg_count_at_A_shorest_dist,
        "ENV_C_avg_count_at_A_shortest_dist": C_avg_count_at_A_shorest_dist,
        "ENV_D_avg_count_at_A_shortest_dist": D_avg_count_at_A_shorest_dist,
        "ENV_A_count_at_B_shortest_dist": A_count_at_B_shortest_dist,
        "ENV_B_count_at_B_shortest_dist": B_count_at_B_shortest_dist,
        "ENV_C_count_at_B_shortest_dist": C_count_at_B_shortest_dist,
        "ENV_D_count_at_B_shortest_dist": D_count_at_B_shortest_dist,
        "ENV_A_avg_count_at_B_shortest_dist": A_avg_count_at_B_shorest_dist,
        "ENV_B_avg_count_at_B_shortest_dist": B_avg_count_at_B_shorest_dist,
        "ENV_C_avg_count_at_B_shortest_dist": C_avg_count_at_B_shorest_dist,
        "ENV_D_avg_count_at_B_shortest_dist": D_avg_count_at_B_shorest_dist,
        "ENV_A_count_at_C_shortest_dist": A_count_at_C_shortest_dist,
        "ENV_B_count_at_C_shortest_dist": B_count_at_C_shortest_dist,
        "ENV_C_count_at_C_shortest_dist": C_count_at_C_shortest_dist,
        "ENV_D_count_at_C_shortest_dist": D_count_at_C_shortest_dist,
        "ENV_A_avg_count_at_C_shortest_dist": A_avg_count_at_C_shorest_dist,
        "ENV_B_avg_count_at_C_shortest_dist": B_avg_count_at_C_shorest_dist,
        "ENV_C_avg_count_at_C_shortest_dist": C_avg_count_at_C_shorest_dist,
        "ENV_D_avg_count_at_C_shortest_dist": D_avg_count_at_C_shorest_dist,
        "ENV_A_count_at_D_shortest_dist": A_count_at_D_shortest_dist,
        "ENV_B_count_at_D_shortest_dist": B_count_at_D_shortest_dist,
        "ENV_C_count_at_D_shortest_dist": C_count_at_D_shortest_dist,
        "ENV_D_count_at_D_shortest_dist": D_count_at_D_shortest_dist,
        "ENV_A_avg_count_at_D_shortest_dist": A_avg_count_at_D_shorest_dist,
        "ENV_B_avg_count_at_D_shortest_dist": B_avg_count_at_D_shorest_dist,
        "ENV_C_avg_count_at_D_shortest_dist": C_avg_count_at_D_shorest_dist,
        "ENV_D_avg_count_at_D_shortest_dist": D_avg_count_at_D_shorest_dist,
    }
    return results



from cifkit import Cif
from cifkit.data import radius_optimization as radius_opt

from SAF.features.interatomic import helper
from SAF.utils import bond, packing


def compute_features(cif: Cif, use_size_constraint: bool, elements):
    A, B, C, D = elements
    A_CIF_rad = cif.radius_values[A]["CIF_radius"]
    B_CIF_rad = cif.radius_values[B]["CIF_radius"]
    C_CIF_rad = cif.radius_values[C]["CIF_radius"]
    D_CIF_rad = cif.radius_values[D]["CIF_radius"]
    CIF_rad_refined, obj_value = radius_opt.get_refined_CIF_radius(
        [A, B, C, D], cif.shortest_bond_pair_distance, elements_ordered=False, use_size_constraint=use_size_constraint
    )
    A_CIF_rad_refined = float(CIF_rad_refined[A])
    B_CIF_rad_refined = float(CIF_rad_refined[B])
    C_CIF_rad_refined = float(CIF_rad_refined[C])
    D_CIF_rad_refined = float(CIF_rad_refined[D])
    min_bond_dists = bond.get_min_distances_by_labels(cif.shortest_bond_pair_distance, [A, B, C, D], labels=["A", "B", "C", "D"])
    # 1) Min bond distances
    AA_dist = min_bond_dists["AA"]
    AB_dist = min_bond_dists["AB"]
    AC_dist = min_bond_dists["AC"]
    AD_dist = min_bond_dists["AD"]
    BB_dist = min_bond_dists["BB"]
    BC_dist = min_bond_dists["BC"]
    BD_dist = min_bond_dists["BD"]
    CC_dist = min_bond_dists["CC"]
    CD_dist = min_bond_dists["CD"]
    DD_dist = min_bond_dists["DD"]
    # 2) Radius from interatomic distances divided by CIF radius (6 of them for quaternary)
    AB_dist_by_AB_sizes = AB_dist / (A_CIF_rad + B_CIF_rad)
    AC_dist_by_AC_sizes = AC_dist / (A_CIF_rad + C_CIF_rad)
    AD_dist_by_AD_sizes = AD_dist / (A_CIF_rad + D_CIF_rad)
    BC_dist_by_BC_sizes = BC_dist / (B_CIF_rad + C_CIF_rad)
    BD_dist_by_BD_sizes = BD_dist / (B_CIF_rad + D_CIF_rad)
    CD_dist_by_CD_sizes = CD_dist / (C_CIF_rad + D_CIF_rad)
    # 3) Radius from interatomic distances divided by 2
    AA_dist_by_2 = AA_dist / 2
    BB_dist_by_2 = BB_dist / 2
    CC_dist_by_2 = CC_dist / 2
    DD_dist_by_2 = DD_dist / 2
    # 4) Refined radius sums (10 of them for quaternary)
    AA_ref_rad_sum = 2 * A_CIF_rad_refined
    BB_ref_rad_sum = 2 * B_CIF_rad_refined
    CC_ref_rad_sum = 2 * C_CIF_rad_refined
    DD_ref_rad_sum = 2 * D_CIF_rad_refined
    AB_ref_rad_sum = A_CIF_rad_refined + B_CIF_rad_refined
    AC_ref_rad_sum = A_CIF_rad_refined + C_CIF_rad_refined
    AD_ref_rad_sum = A_CIF_rad_refined + D_CIF_rad_refined
    BC_ref_rad_sum = B_CIF_rad_refined + C_CIF_rad_refined
    BD_ref_rad_sum = B_CIF_rad_refined + D_CIF_rad_refined
    CD_ref_rad_sum = C_CIF_rad_refined + D_CIF_rad_refined
    # 5) Percent differences between refined and CIF radii
    percent_diff_A = (A_CIF_rad_refined - A_CIF_rad) / A_CIF_rad
    percent_diff_B = (B_CIF_rad_refined - B_CIF_rad) / B_CIF_rad
    percent_diff_C = (C_CIF_rad_refined - C_CIF_rad) / C_CIF_rad
    percent_diff_D = (D_CIF_rad_refined - D_CIF_rad) / D_CIF_rad
    # 6) Size ratios of CIF radii
    Asize_by_Bsize = A_CIF_rad / B_CIF_rad
    Bsize_by_Csize = B_CIF_rad / C_CIF_rad
    Csize_by_Dsize = C_CIF_rad / D_CIF_rad
    Asize_by_Csize = A_CIF_rad / C_CIF_rad
    Asize_by_Dsize = A_CIF_rad / D_CIF_rad
    Bsize_by_Dsize = B_CIF_rad / D_CIF_rad
    # 7) Radius from interatomic distances divided by CIF radius
    AA_dist_by_2_by_Asize = AA_dist_by_2 / A_CIF_rad
    BB_dist_by_2_by_Bsize = BB_dist_by_2 / B_CIF_rad
    CC_dist_by_2_by_Csize = CC_dist_by_2 / C_CIF_rad
    DD_dist_by_2_by_Dsize = DD_dist_by_2 / D_CIF_rad
    # 8) Difference between the iteratomic distance and the expected refined radius sum
    AA_minus_ref_diff = (AA_dist - AA_ref_rad_sum) / AA_dist
    BB_minus_ref_diff = (BB_dist - BB_ref_rad_sum) / BB_dist
    CC_minus_ref_diff = (CC_dist - CC_ref_rad_sum) / CC_dist
    DD_minus_ref_diff = (DD_dist - DD_ref_rad_sum) / DD_dist
    AB_minus_ref_diff = (AB_dist - AB_ref_rad_sum) / AB_dist
    AC_minus_ref_diff = (AC_dist - AC_ref_rad_sum) / AC_dist
    AD_minus_ref_diff = (AD_dist - AD_ref_rad_sum) / AD_dist
    BC_minus_ref_diff = (BC_dist - BC_ref_rad_sum) / BC_dist
    BD_minus_ref_diff = (BD_dist - BD_ref_rad_sum) / BD_dist
    CD_minus_ref_diff = (CD_dist - CD_ref_rad_sum) / CD_dist
    # 9) Get the shortest homo/hetroatomic distance
    homoatomic_dists = helper.get_bond_distances_by_labels(min_bond_dists, ["AA", "BB", "CC", "DD"])
    heteroatomic_dists = helper.get_bond_distances_by_labels(min_bond_dists, ["AB", "AC", "AD", "BC", "BD", "CD"])
    shortest_heteroatomic_dist = helper.get_min_from_dists(heteroatomic_dists)
    shortest_homoatomic_dist = helper.get_min_from_dists(homoatomic_dists)
    # Parse the CIF radius based on the shortest homo/heteroatomic dist
    rad_dict = {
        "AA": A_CIF_rad,
        "BB": B_CIF_rad,
        "CC": C_CIF_rad,
        "DD": D_CIF_rad,
        "AB": A_CIF_rad + B_CIF_rad,
        "AC": A_CIF_rad + C_CIF_rad,
        "AD": A_CIF_rad + D_CIF_rad,
        "BC": B_CIF_rad + C_CIF_rad,
        "BD": B_CIF_rad + D_CIF_rad,
        "CD": C_CIF_rad + D_CIF_rad,
    }

    refined_rad_dict = {
        "AA": A_CIF_rad_refined,
        "BB": B_CIF_rad_refined,
        "CC": C_CIF_rad_refined,
        "DD": D_CIF_rad_refined,
        "AB": A_CIF_rad_refined + B_CIF_rad_refined,
        "AC": A_CIF_rad_refined + C_CIF_rad_refined,
        "AD": A_CIF_rad_refined + D_CIF_rad_refined,
        "BC": B_CIF_rad_refined + C_CIF_rad_refined,
        "BD": B_CIF_rad_refined + D_CIF_rad_refined,
        "CD": C_CIF_rad_refined + D_CIF_rad_refined,
    }

    percent_diffs = [percent_diff_A, percent_diff_B, percent_diff_C, percent_diff_D]
    shortest_homo_key = helper.get_shortest_homo_key(min_bond_dists, shortest_homoatomic_dist)
    shortest_hetero_key = helper.get_shortest_hetero_key(min_bond_dists, shortest_heteroatomic_dist)

    uni_features = helper.get_shortest_homo_heteroatomic_features(
        shortest_homoatomic_dist,
        shortest_homo_key,
        shortest_heteroatomic_dist,
        shortest_hetero_key,
        rad_dict,
        refined_rad_dict,
    )

    highest_refined_percent_diff = helper.get_highest_refined_percent_diff(percent_diffs)
    lowest_refined_percent_diff = helper.get_lowest_refined_percent_diff(percent_diffs)

    # Packing efficiency
    packing_efficiency = packing.compute_efficiency(
        cif,
        {
            A: A_CIF_rad_refined,
            B: B_CIF_rad_refined,
            C: C_CIF_rad_refined,
            D: D_CIF_rad_refined,
        },
    )

    # Distances
    data = {
        "Entry": cif.file_name_without_ext,
        "Formula": cif.formula,
        "Structure": cif.structure,
        "A": A,
        "B": B,
        "C": C,
        "D": D,
        # 1. Shortest distances
        "INT_AA_dist": AA_dist,
        "INT_BB_dist": BB_dist,
        "INT_CC_dist": CC_dist,
        "INT_DD_dist": DD_dist,
        "INT_AB_dist": AB_dist,
        "INT_AC_dist": AC_dist,
        "INT_AD_dist": AD_dist,
        "INT_BC_dist": BC_dist,
        "INT_BD_dist": BD_dist,
        "INT_CD_dist": CD_dist,
        # 2. CIF radius
        "INT_Asize": A_CIF_rad,
        "INT_Bsize": B_CIF_rad,
        "INT_Csize": C_CIF_rad,
        "INT_Dsize": D_CIF_rad,
        # 3. CIF radius ratio
        "INT_Asize_by_Bsize": Asize_by_Bsize,
        "INT_Bsize_by_Csize": Bsize_by_Csize,
        "INT_Csize_by_Dsize": Csize_by_Dsize,
        "INT_Asize_by_Csize": Asize_by_Csize,
        "INT_Asize_by_Dsize": Asize_by_Dsize,
        "INT_Bsize_by_Dsize": Bsize_by_Dsize,
        # 4. Radius from distance divided by CIF radius
        "INT_AA_dist_by_2_by_Asize": AA_dist_by_2_by_Asize,
        "INT_BB_dist_by_2_by_Bsize": BB_dist_by_2_by_Bsize,
        "INT_CC_dist_by_2_by_Csize": CC_dist_by_2_by_Csize,
        "INT_DD_dist_by_2_by_Dsize": DD_dist_by_2_by_Dsize,
        "INT_AB_dist_by_ABsizes": AB_dist_by_AB_sizes,
        "INT_AC_dist_by_ACsizes": AC_dist_by_AC_sizes,
        "INT_AD_dist_by_ADsizes": AD_dist_by_AD_sizes,
        "INT_BC_dist_by_BCsizes": BC_dist_by_BC_sizes,
        "INT_BD_dist_by_BDsizes": BD_dist_by_BD_sizes,
        "INT_CD_dist_by_CDsizes": CD_dist_by_CD_sizes,
        # 5. Refined CIF radius
        "INT_Asize_ref": A_CIF_rad_refined,
        "INT_Bsize_ref": B_CIF_rad_refined,
        "INT_Csize_ref": C_CIF_rad_refined,
        "INT_Dsize_ref": D_CIF_rad_refined,
        # 6. Difference between CIF radius and refined CIF radius
        "INT_percent_diff_A_by_100": percent_diff_A,
        "INT_percent_diff_B_by_100": percent_diff_B,
        "INT_percent_diff_C_by_100": percent_diff_C,
        "INT_percent_diff_D_by_100": percent_diff_D,
        # 7. Difference between interatomic an_distce and refined radius sum
        "INT_AA_minus_ref_diff": AA_minus_ref_diff,
        "INT_BB_minus_ref_diff": BB_minus_ref_diff,
        "INT_CC_minus_ref_diff": CC_minus_ref_diff,
        "INT_DD_minus_ref_diff": DD_minus_ref_diff,
        "INT_AB_minus_ref_diff": AB_minus_ref_diff,
        "INT_AC_minus_ref_diff": AC_minus_ref_diff,
        "INT_AD_minus_ref_diff": AD_minus_ref_diff,
        "INT_BC_minus_ref_diff": BC_minus_ref_diff,
        "INT_BD_minus_ref_diff": BD_minus_ref_diff,
        "INT_CD_minus_ref_diff": CD_minus_ref_diff,
        # 8. Objective value of the radius optimization
        "INT_R_factor": obj_value,
        # 9. Universal features
        "INT_UNI_shortest_homoatomic_dist": shortest_homoatomic_dist,
        "INT_UNI_shortest_heteroatomic_dist": shortest_heteroatomic_dist,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_atom_size"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": uni_features["shortest_heteroatomic_dist_by_sum_of_atom_sizes"],
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_refined_atom_sizes"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": uni_features["shortest_heteroatomic_dist_by_refined_atom_sizes"],
        "INT_UNI_highest_refined_percent_diff_abs": highest_refined_percent_diff,
        "INT_UNI_lowest_refined_percent_diff_abs": lowest_refined_percent_diff,
        "INT_UNI_refined_packing_efficiency": packing_efficiency,
    }

    uni_data = {
        "Entry": cif.file_name_without_ext,
        "Formula": cif.formula,
        "INT_UNI_shortest_homoatomic_dist": shortest_homoatomic_dist,
        "INT_UNI_shortest_heteroatomic_dist": shortest_heteroatomic_dist,
        "INT_UNI_shortest_homoatomic_dist_by_2_by_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_atom_size"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_atom_sizes": uni_features["shortest_heteroatomic_dist_by_sum_of_atom_sizes"],
        "INT_UNI_shortest_homoatomic_dist_by_2_by_refined_atom_size": uni_features["shortest_homoatomic_dist_by_2_by_refined_atom_sizes"],
        "INT_UNI_shortest_heteroatomic_dist_by_sum_of_refined_atom_sizes": uni_features["shortest_heteroatomic_dist_by_refined_atom_sizes"],
        "INT_UNI_highest_refined_percent_diff_abs": highest_refined_percent_diff,
        "INT_UNI_lowest_refined_percent_diff_abs": lowest_refined_percent_diff,
        "INT_UNI_refined_packing_efficiency": packing_efficiency,
    }
    return data, uni_data



from cifkit import Cif

from SAF.features.environment.ternary_helper import (
    compute_avg_homoatomic_dist_by_site_shortest_dist,
    compute_homoatomic_dist_by_site_shortest_dist,
    get_avg_R_and_M_and_X_count_in_per_element,
    get_R_and_M_and_X_count_in_best_label_per_element,
)
from SAF.features.environment.util import (
    count_first_second_min_dist,
    extract_avg_shortest_dist_with_tol,
    extract_best_labels,
    extract_shortest_dist_with_tol,
    get_avg_second_by_first_shortest_dist_ratio,
)


def compute_features(cif: Cif, elements):
    connections = cif.connections
    R, M, X = elements

    (
        R_avg_homoatomic_dist_by_shortest_dist,
        M_avg_homoatomic_dist_by_shortest_dist,
        X_avg_homoatomic_dist_by_shortest_dist,
    ) = compute_avg_homoatomic_dist_by_site_shortest_dist(connections, R, M, X)
    (
        R_count_at_R_shortest_dist,
        R_count_at_M_shortest_dist,
        R_count_at_X_shortest_dist,
        M_count_at_R_shortest_dist,
        M_count_at_M_shortest_dist,
        M_count_at_X_shortest_dist,
        X_count_at_R_shortest_dist,
        X_count_at_M_shortest_dist,
        X_count_at_X_shortest_dist,
    ) = get_R_and_M_and_X_count_in_best_label_per_element(connections, R, M, X)
    (
        R_avg_count_at_R_shorest_dist,
        R_avg_count_at_M_shorest_dist,
        R_avg_count_at_X_shorest_dist,
        M_avg_count_at_R_shorest_dist,
        M_avg_count_at_M_shorest_dist,
        M_avg_count_at_X_shorest_dist,
        X_avg_count_at_R_shorest_dist,
        X_avg_count_at_M_shorest_dist,
        X_avg_count_at_X_shorest_dist,
    ) = get_avg_R_and_M_and_X_count_in_per_element(connections, R, M, X)
    # 1. Compute the first, second shortest distances and count
    first_second_dist_per_site_data = count_first_second_min_dist(connections)
    # 2. Compute the best site for each label, determined by the shortest distance
    best_site_data = extract_best_labels(first_second_dist_per_site_data)
    # 3. Tol result
    tol_results = extract_shortest_dist_with_tol(best_site_data, connections)
    R_shortest_dist_count_within_tol = tol_results[R]["shortest_dist_count_within_tol"]
    M_shortest_dist_count_within_tol = tol_results[M]["shortest_dist_count_within_tol"]
    X_shortest_dist_count_within_tol = tol_results[X]["shortest_dist_count_within_tol"]
    avg_tol_results = extract_avg_shortest_dist_with_tol(connections)
    R_avg_shortest_dist_within_tol_count = avg_tol_results[R]["avg_shortest_dist_within_tol_count"]
    M_avg_shortest_dist_within_tol_count = avg_tol_results[M]["avg_shortest_dist_within_tol_count"]
    X_avg_shortest_dist_within_tol_count = avg_tol_results[X]["avg_shortest_dist_within_tol_count"]
    R_best_label = best_site_data[R]["best_label"]
    M_best_label = best_site_data[M]["best_label"]
    X_best_label = best_site_data[X]["best_label"]
    (
        R_homoatomic_dist_by_shortest_dist,
        M_homoatomic_dist_by_shortest_dist,
        X_homoatomic_dist_by_shortest_dist,
    ) = compute_homoatomic_dist_by_site_shortest_dist(connections, R_best_label, M_best_label, X_best_label)
    # First shortest distance
    R_shortest_dist = best_site_data[R]["best_label_details"]["shortest_dist"]
    M_shortest_dist = best_site_data[M]["best_label_details"]["shortest_dist"]
    X_shortest_dist = best_site_data[X]["best_label_details"]["shortest_dist"]
    # Second shortest distance
    R_second_shortest_dist = best_site_data[R]["best_label_details"]["second_shortest_dist"]
    M_second_shortest_dist = best_site_data[M]["best_label_details"]["second_shortest_dist"]
    X_second_shortest_dist = best_site_data[X]["best_label_details"]["second_shortest_dist"]
    # First shortest distance count
    R_shortest_dist_count = best_site_data[R]["best_label_details"]["counts"][R_shortest_dist]
    M_shortest_dist_count = best_site_data[M]["best_label_details"]["counts"][M_shortest_dist]
    X_shortest_dist_count = best_site_data[X]["best_label_details"]["counts"][X_shortest_dist]
    # Second shorest distance count
    R_second_shortest_dist_count = best_site_data[R]["best_label_details"]["counts"][R_second_shortest_dist]
    M_second_shortest_dist_count = best_site_data[M]["best_label_details"]["counts"][M_second_shortest_dist]
    X_second_shortest_dist_count = best_site_data[X]["best_label_details"]["counts"][X_second_shortest_dist]
    # Avg shortest distance count across site labels per element
    R_avg_shortest_dist_count = best_site_data[R]["avg_shortest_dist"]
    M_avg_shortest_dist_count = best_site_data[M]["avg_shortest_dist"]
    X_avg_shortest_dist_count = best_site_data[X]["avg_shortest_dist"]
    # Avg second shortest distance count across site labels per element
    R_avg_second_shortest_dist_count = best_site_data[R]["avg_second_shortest_dist_count"]
    M_avg_second_shortest_dist_count = best_site_data[M]["avg_second_shortest_dist_count"]
    X_avg_second_shortest_dist_count = best_site_data[M]["avg_second_shortest_dist_count"]
    # Get avg second by first shortest distance ratio across site labels per element
    avg_second_by_first_dist = get_avg_second_by_first_shortest_dist_ratio(first_second_dist_per_site_data, connections)
    R_avg_second_by_first_shortest_dist = avg_second_by_first_dist[R]["avg_second_by_first_shortest_dist"]
    M_avg_second_by_first_shortest_dist = avg_second_by_first_dist[M]["avg_second_by_first_shortest_dist"]
    X_avg_second_by_first_shortest_dist = avg_second_by_first_dist[X]["avg_second_by_first_shortest_dist"]

    results = {
        # Shortest distance count
        "ENV_R_shortest_dist_count": R_shortest_dist_count,
        "ENV_M_shortest_dist_count": M_shortest_dist_count,
        "ENV_X_shortest_dist_count": X_shortest_dist_count,
        # Avg shortest distance count
        "ENV_R_avg_shortest_dist_count": R_avg_shortest_dist_count,
        "ENV_M_avg_shortest_dist_count": M_avg_shortest_dist_count,
        "ENV_X_avg_shortest_dist_count": X_avg_shortest_dist_count,
        # Shortest distance count within tolerance
        "ENV_R_shortest_tol_dist_count": R_shortest_dist_count_within_tol,
        "ENV_M_shortest_tol_dist_count": M_shortest_dist_count_within_tol,
        "ENV_X_shortest_tol_dist_count": X_shortest_dist_count_within_tol,
        # Avg shortest distance count within tolerance
        "ENV_R_avg_shortest_dist_within_tol_count": R_avg_shortest_dist_within_tol_count,
        "ENV_M_avg_shortest_dist_within_tol_count": M_avg_shortest_dist_within_tol_count,
        "ENV_X_avg_shortest_dist_within_tol_count": X_avg_shortest_dist_within_tol_count,
        # Second by first shortest distance ratio
        "ENV_R_second_by_first_shortest_dist": R_second_shortest_dist / R_shortest_dist,
        "ENV_M_second_by_first_shortest_dist": M_second_shortest_dist / M_shortest_dist,
        "ENV_X_second_by_first_shortest_dist": X_second_shortest_dist / X_shortest_dist,
        # Avg second by first shortest distance ratio
        "ENV_R_avg_second_by_first_shortest_dist": R_avg_second_by_first_shortest_dist,
        "ENV_M_avg_second_by_first_shortest_dist": M_avg_second_by_first_shortest_dist,
        "ENV_X_avg_second_by_first_shortest_dist": X_avg_second_by_first_shortest_dist,
        # Second shortest distance count
        "ENV_R_second_shortest_dist_count": R_second_shortest_dist_count,
        "ENV_M_second_shortest_dist_count": M_second_shortest_dist_count,
        "ENV_X_second_shortest_dist_count": X_second_shortest_dist_count,
        # Avg second shortest distance count
        "ENV_R_avg_second_shortest_dist_count": R_avg_second_shortest_dist_count,
        "ENV_M_avg_second_shortest_dist_count": M_avg_second_shortest_dist_count,
        "ENV_X_avg_second_shortest_dist_count": X_avg_second_shortest_dist_count,
        # Homoatomic distance by shortest distance
        "ENV_R_homoatomic_dist_by_shortest_dist": R_homoatomic_dist_by_shortest_dist,
        "ENV_M_homoatomic_dist_by_shortest_dist": M_homoatomic_dist_by_shortest_dist,
        "ENV_X_homoatomic_dist_by_shortest_dist": X_homoatomic_dist_by_shortest_dist,
        # Avg homoatomic distance by shortest distance
        "ENV_R_avg_homoatomic_dist_by_shortest_dist": R_avg_homoatomic_dist_by_shortest_dist,
        "ENV_M_avg_homoatomic_dist_by_shortest_dist": M_avg_homoatomic_dist_by_shortest_dist,
        "ENV_X_avg_homoatomic_dist_by_shortest_dist": X_avg_homoatomic_dist_by_shortest_dist,
        # Count at shortest distance
        "ENV_R_count_at_R_shortest_dist": R_count_at_R_shortest_dist,
        "ENV_M_count_at_R_shortest_dist": M_count_at_R_shortest_dist,
        "ENV_X_count_at_R_shortest_dist": X_count_at_R_shortest_dist,
        "ENV_R_avg_count_at_R_shortest_dist": R_avg_count_at_R_shorest_dist,
        "ENV_M_avg_count_at_R_shortest_dist": M_avg_count_at_R_shorest_dist,
        "ENV_X_avg_count_at_R_shortest_dist": X_avg_count_at_R_shorest_dist,
        "ENV_R_count_at_M_shortest_dist": R_count_at_M_shortest_dist,
        "ENV_M_count_at_M_shortest_dist": M_count_at_M_shortest_dist,
        "ENV_X_count_at_M_shortest_dist": X_count_at_M_shortest_dist,
        "ENV_R_avg_count_at_M_shortest_dist": R_avg_count_at_M_shorest_dist,
        "ENV_M_avg_count_at_M_shortest_dist": M_avg_count_at_M_shorest_dist,
        "ENV_X_avg_count_at_M_shortest_dist": X_avg_count_at_M_shorest_dist,
        "ENV_R_count_at_X_shortest_dist": R_count_at_X_shortest_dist,
        "ENV_M_count_at_X_shortest_dist": M_count_at_X_shortest_dist,
        "ENV_X_count_at_X_shortest_dist": X_count_at_X_shortest_dist,
        "ENV_R_avg_count_at_X_shortest_dist": R_avg_count_at_X_shorest_dist,
        "ENV_M_avg_count_at_X_shortest_dist": M_avg_count_at_X_shorest_dist,
        "ENV_X_avg_count_at_X_shortest_dist": X_avg_count_at_X_shorest_dist,
    }
    return results



from cifkit import Cif
from cifkit.coordination.geometry import compute_polyhedron_metrics
from cifkit.utils import string_parser
from scipy.spatial import ConvexHull


def get_CN_metrics_per_method(cif: Cif):
    """Find the best polyhedron for each label based on the minimum
    distance between the reference atom to the average position of
    connected atoms."""
    max_gaps_per_label = cif.CN_max_gap_per_site
    connections = cif.connections
    site_data = {}
    for label, CN_data_per_method in max_gaps_per_label.items():
        site_data[label] = {}
        for method, CN_data in CN_data_per_method.items():
            connection_data = connections[label][: CN_data["CN"]]
            polyhedron_points = []
            # Only if there are 4 or more points in the polyhedron
            if len(connection_data) > 3:
                for connection in connection_data:
                    polyhedron_points.append(connection[3])
            else:
                continue
            # Add the central atom as the last element
            polyhedron_points.append(connection_data[0][2])
            # Try to make a polyhedron
            try:
                hull = ConvexHull(polyhedron_points)
            except Exception:
                print(f"Error in determining polyhedron for {label} using {method} - skipped")
                site_data[label][method] = None
                continue  # Move to the next method
            # Returns none if there is any error
            polyhedron_metrics = compute_polyhedron_metrics(polyhedron_points, hull)
            site_data[label][method] = polyhedron_metrics
    return site_data


def _compute_number_of_atoms_in_binary_CN(label_connections, CN_metrics, A, B):
    """Compute the number of A, B atoms for each label, and 4 methods
    per label."""
    CN_atom_count_data = {}
    for site_label, method_data in CN_metrics.items():
        CN_atom_count_data[site_label] = {}
        for method, data in method_data.items():
            # Get the number of CN per method
            CN = data["number_of_vertices"]
            CN_connections = label_connections[site_label][:CN]
            A_element_count = 0
            B_element_count = 0
            for connection in CN_connections:
                connected_label = connection[0]
                parsed_label = string_parser.get_atom_type_from_label(connected_label)
                if parsed_label == A:
                    A_element_count += 1
                elif parsed_label == B:
                    B_element_count += 1
            CN_atom_count_data[site_label][method] = {
                "A_count": A_element_count,
                "B_count": B_element_count,
            }
    return CN_atom_count_data


def _compute_number_of_atoms_in_ternary_CN(label_connections, CN_metrics, R, M, X):
    """Compute the number of R, M, X atoms for each label, and 4 methods
    per label."""
    CN_atom_count_data = {}
    for site_label, method_data in CN_metrics.items():
        CN_atom_count_data[site_label] = {}
        for method, data in method_data.items():
            # Get the number of CN per method
            CN = data["number_of_vertices"]
            CN_connections = label_connections[site_label][:CN]
            R_element_count = 0
            M_element_count = 0
            X_element_count = 0
            for connection in CN_connections:
                connected_label = connection[0]
                parsed_label = string_parser.get_atom_type_from_label(connected_label)
                if parsed_label == R:
                    R_element_count += 1
                elif parsed_label == M:
                    M_element_count += 1
                elif parsed_label == X:
                    X_element_count += 1
            CN_atom_count_data[site_label][method] = {
                "R_count": R_element_count,
                "M_count": M_element_count,
                "X_count": X_element_count,
            }
    return CN_atom_count_data


def _compute_number_of_atoms_in_quaternary_CN(label_connections, CN_metrics, A, B, C, D):
    """Compute the number of A, B, C, D atoms for each label, and 4
    methods per label."""
    CN_atom_count_data = {}
    for site_label, method_data in CN_metrics.items():
        CN_atom_count_data[site_label] = {}
        for method, data in method_data.items():
            # Get the number of CN per method
            CN = data["number_of_vertices"]
            CN_connections = label_connections[site_label][:CN]
            A_element_count = 0
            B_element_count = 0
            C_element_count = 0
            D_element_count = 0
            for connection in CN_connections:
                connected_label = connection[0]
                parsed_label = string_parser.get_atom_type_from_label(connected_label)
                if parsed_label == A:
                    A_element_count += 1
                elif parsed_label == B:
                    B_element_count += 1
                elif parsed_label == C:
                    C_element_count += 1
                elif parsed_label == D:
                    D_element_count += 1
            CN_atom_count_data[site_label][method] = {
                "A_count": A_element_count,
                "B_count": B_element_count,
                "C_count": C_element_count,
                "D_count": D_element_count,
            }
    return CN_atom_count_data


def _compute_min_max_avg_per_atomic_label(site_data):
    """Calculate the minimum, maximum, and average of the 4 methods for
    each atomic site label, e.g., "Sb1", "Sb2"."""
    min_max_avg_per_site = {}
    # Iterate over each element (like Sb1, Th1, etc.)
    for label in site_data:
        metrics = {}
        # Iterate over each method under the element
        for method in site_data[label]:
            # Collect data for each metric within the method
            for key in site_data[label][method]:
                if key not in metrics:
                    metrics[key] = []
                metrics[key].append(site_data[label][method][key])
        # Calculate min, max, and avg for each metric within the site label
        min_max_avg_per_site[label] = {}
        for key in metrics:
            min_max_avg_per_site[label][key] = {
                "min": min(metrics[key]),
                "max": max(metrics[key]),
                "avg": sum(metrics[key]) / len(metrics[key]),
            }
    return min_max_avg_per_site


def _compute_global_avg_for_min_max_avg_metrics(min_max_avg_result):
    """Calculate global averages of all minimums, maximums, and averages
    across all labels from a pre-computed min-max-avg result.

    From the above function, we copmuted the min, max, and avg for each
    label, now we compute the global averages across all labels. This is
    important for ML since we are trying capture the global min, max,
    and avg of the coordination environment values.
    """
    global_sums = {"min": {}, "max": {}, "avg": {}}
    global_counts = {"min": {}, "max": {}, "avg": {}}
    # Summing values across all labels
    for label_data in min_max_avg_result.values():
        for metric, values in label_data.items():
            for stat in ["min", "max", "avg"]:
                if metric not in global_sums[stat]:
                    global_sums[stat][metric] = 0
                    global_counts[stat][metric] = 0
                global_sums[stat][metric] += values[stat]
                global_counts[stat][metric] += 1
    # Calculating global averages
    global_avg = {"min": {}, "max": {}, "avg": {}}
    for stat in ["min", "max", "avg"]:
        for metric, sum_value in global_sums[stat].items():
            global_avg[stat][metric] = float(sum_value / global_counts[stat][metric])
    return global_avg


def get_CN_atom_count_data(cif: Cif, elements):
    CN_metrics = get_CN_metrics_per_method(cif)
    min_max_avg_CN_metrics = _compute_min_max_avg_per_atomic_label(CN_metrics)
    if len(cif.unique_elements) == 2:
        A = elements[0]
        B = elements[1]
        CN_atom_count_data = _compute_number_of_atoms_in_binary_CN(cif.connections, CN_metrics, A, B)
    if len(cif.unique_elements) == 3:
        R = elements[0]
        M = elements[1]
        X = elements[2]
        CN_atom_count_data = _compute_number_of_atoms_in_ternary_CN(cif.connections, CN_metrics, R, M, X)
    if len(cif.unique_elements) == 4:
        A = elements[0]
        B = elements[1]
        C = elements[2]
        D = elements[3]
        CN_atom_count_data = _compute_number_of_atoms_in_quaternary_CN(cif.connections, CN_metrics, A, B, C, D)
    # Compute the min, max, and avg for each label. Each label has 4 methods
    min_max_avg_CN_count = _compute_min_max_avg_per_atomic_label(CN_atom_count_data)
    # Find the average of the min, max, and avg from
    avg_CN_metrics = _compute_global_avg_for_min_max_avg_metrics(min_max_avg_CN_metrics)
    avg_CN_atom_count = _compute_global_avg_for_min_max_avg_metrics(min_max_avg_CN_count)
    return avg_CN_metrics, avg_CN_atom_count
